import logging

from django.contrib.contenttypes.models import ContentType
from django.db.models.signals import post_delete
from django.dispatch import receiver

from apis_core.apis_metainfo.models import Uri

logger = logging.getLogger(__name__)


@receiver(post_delete)
def remove_stale_uris(sender, instance, *args, **kwargs):
    content_type = ContentType.objects.get_for_model(instance)
    if isinstance(instance.pk, int):
        uris = Uri.objects.filter(content_type=content_type, object_id=instance.pk)
        for uri in uris:
            logger.info(
                "Deleting uri %s as a result of deleting %s", repr(uri), repr(instance)
            )
            uri.delete()



import logging

from django.contrib.contenttypes.models import ContentType
from django.db.models import BooleanField, CharField, TextField
from django.db.models.fields.related import ForeignKey, ManyToManyField
from django.db.models.query import QuerySet
from django.forms import model_to_dict
from django.urls import reverse

from apis_core.generic.helpers import mro_paths, permission_fullname
from apis_core.generic.signals import (
    post_duplicate,
    post_merge_with,
    pre_duplicate,
    pre_merge_with,
)

logger = logging.getLogger(__name__)


class GenericModel:
    def __repr__(self):
        if id := getattr(self, "id", None):
            return super().__repr__() + f" (ID: {id})"
        return super().__repr__()

    @classmethod
    def get_listview_url(cls):
        ct = ContentType.objects.get_for_model(cls)
        return reverse("apis_core:generic:list", args=[ct])

    @classmethod
    def get_createview_url(cls):
        ct = ContentType.objects.get_for_model(cls)
        return reverse("apis_core:generic:create", args=[ct])

    @classmethod
    def get_importview_url(cls):
        ct = ContentType.objects.get_for_model(cls)
        return reverse("apis_core:generic:import", args=[ct])

    @classmethod
    def get_openapi_tags(cls):
        return [item[-1] for item in mro_paths(cls)]

    def get_edit_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:update", args=[ct, self.id])

    def get_enrich_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:enrich", args=[ct, self.id])

    def get_absolute_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:detail", args=[ct, self.id])

    def get_delete_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:delete", args=[ct, self.id])

    def get_merge_url(self, other_id):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:merge", args=[ct, self.id, other_id])

    def get_select_merge_or_enrich_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:selectmergeorenrich", args=[ct, self.id])

    def get_create_success_url(self):
        return self.get_absolute_url()

    def get_update_success_url(self):
        return self.get_edit_url()

    def get_api_detail_endpoint(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:genericmodelapi-detail", args=[ct, self.id])

    @classmethod
    def get_change_permission(self):
        return permission_fullname("change", self)

    @classmethod
    def get_add_permission(self):
        return permission_fullname("add", self)

    @classmethod
    def get_delete_permission(self):
        return permission_fullname("delete", self)

    def get_merge_charfield_value(self, other: CharField, field: CharField):
        res = getattr(self, field.name)
        if not field.choices:
            otherres = getattr(other, field.name, res)
            if otherres != res:
                res += f" ({otherres})"
        return res

    def get_merge_textfield_value(self, other: TextField, field: TextField):
        res = getattr(self, field.name)
        if getattr(other, field.name):
            res += "\n" + f"Merged from {other}:\n" + getattr(other, field.name)
        return res

    def get_merge_booleanfield(self, other: BooleanField, field: BooleanField):
        return getattr(other, field.name)

    def get_field_value_after_merge(self, other, field):
        """
        This method finds the value of a field after merging `other` into `self`.
        It first tries to find a merge method that is specific to that field
        (merge_{fieldname}) and then tries to find a method that is specific to
        the type of the field (merge_{fieldtype})
        If neither of those exist, it uses the others field value if the field
        in self is not set, otherwise it keeps the value in self.
        """
        fieldtype = field.get_internal_type().lower()
        # if there is a `get_merge_{fieldname}` method in this model, use that one
        if callable(getattr(self, f"get_merge_{field.name}_value", None)):
            return getattr(self, f"get_merge_{field.name}_value")(other)
        # otherwise we check if there is a method for the field type and use that one
        elif callable(getattr(self, f"get_merge_{fieldtype}_value", None)):
            return getattr(self, f"get_merge_{fieldtype}_value")(other, field)
        else:
            if not getattr(self, field.name):
                return getattr(other, field.name)
        return getattr(self, field.name)

    def merge_fields(self, other):
        """
        This method iterates through the model fields and uses the
        `get_field_value_after_merge` method to copy values from `other` to `self`.
        It is called by the `merge_with` method.
        """
        for field in self._meta.fields:
            newval = self.get_field_value_after_merge(other, field)
            if newval != getattr(self, field.name):
                setattr(self, field.name, newval)
        self.save()

    def merge_with(self, entities):
        if self in entities:
            entities.remove(self)
        origin = self.__class__
        pre_merge_with.send(sender=origin, instance=self, entities=entities)

        # TODO: check if these imports can be put to top of module without
        #  causing circular import issues.
        from apis_core.apis_metainfo.models import Uri

        e_a = type(self).__name__
        self_model_class = ContentType.objects.get(model__iexact=e_a).model_class()
        if isinstance(entities, int):
            entities = self_model_class.objects.get(pk=entities)
        if not isinstance(entities, list) and not isinstance(entities, QuerySet):
            entities = [entities]
            entities = [
                self_model_class.objects.get(pk=ent) if isinstance(ent, int) else ent
                for ent in entities
            ]
        for ent in entities:
            e_b = type(ent).__name__
            if e_a != e_b:
                continue
            for f in ent._meta.local_many_to_many:
                if not f.name.endswith("_set"):
                    sl = list(getattr(self, f.name).all())
                    for s in getattr(ent, f.name).all():
                        if s not in sl:
                            getattr(self, f.name).add(s)
            self_content_type = ContentType.objects.get_for_model(self)
            ent_content_type = ContentType.objects.get_for_model(ent)
            Uri.objects.filter(content_type=ent_content_type, object_id=ent.id).update(
                content_type=self_content_type, object_id=self.id
            )

        for ent in entities:
            self.merge_fields(ent)

        post_merge_with.send(sender=origin, instance=self, entities=entities)

        for ent in entities:
            ent.delete()

    def duplicate(self):
        origin = self.__class__
        pre_duplicate.send(sender=origin, instance=self)
        # usually, copying instances would work like
        # https://docs.djangoproject.com/en/4.2/topics/db/queries/#copying-model-instances
        # but we are working with abstract classes,
        # so we have to do it by hand  using model_to_dict:(
        objdict = model_to_dict(self)

        # remove unique fields from dict representation
        unique_fields = [field for field in self._meta.fields if field.unique]
        for field in unique_fields:
            logger.info(f"Duplicating {self}: ignoring unique field {field.name}")
            objdict.pop(field.name, None)

        # remove related fields from dict representation
        related_fields = [
            field for field in self._meta.get_fields() if field.is_relation
        ]
        for field in related_fields:
            objdict.pop(field.name, None)

        newobj = type(self).objects.create(**objdict)

        for field in related_fields:
            # we are not using `isinstance` because we want to
            # differentiate between different levels of inheritance
            if type(field) is ForeignKey:
                setattr(newobj, field.name, getattr(self, field.name))
            if type(field) is ManyToManyField:
                objfield = getattr(newobj, field.name)
                values = getattr(self, field.name).all()
                objfield.set(values)

        newobj.save()
        post_duplicate.send(sender=origin, instance=self, duplicate=newobj)
        return newobj

    duplicate.alters_data = True

    def uri_set(self):
        ct = ContentType.objects.get_for_model(self)
        return (
            ContentType.objects.get(app_label="apis_metainfo", model="uri")
            .model_class()
            .objects.filter(content_type=ct, object_id=self.id)
            .all()
        )



from django import template
from django.contrib.contenttypes.models import ContentType

from apis_core.core.templatetags.core import get_model_fields
from apis_core.generic.abc import GenericModel
from apis_core.generic.helpers import template_names_via_mro

register = template.Library()


@register.filter
def contenttype(model):
    return ContentType.objects.get_for_model(model)


@register.simple_tag
def modeldict(instance, fields=None, exclude=None):
    data = {}
    for f in get_model_fields(instance):
        if not getattr(f, "editable", False):
            continue
        if fields is not None and f.name not in fields:
            continue
        if exclude and f.name in exclude:
            continue
        field = instance._meta.get_field(f.name)
        data[field] = instance._get_FIELD_display(field)
        if getattr(field, "remote_field", False):
            data[field] = getattr(instance, field.name)
        if getattr(field, "m2m_field_name", False):
            values = getattr(instance, field.name).all()
            data[field] = ", ".join([str(value) for value in values])
    return data


@register.simple_tag
def contenttypes(app_labels=None):
    if app_labels:
        app_labels = app_labels.split(",")
        return ContentType.objects.filter(app_label__in=app_labels)
    return ContentType.objects.all()


def is_genericmodel(content_type: ContentType):
    model_class = content_type.model_class()
    return model_class is not None and issubclass(model_class, GenericModel)


@register.simple_tag
def genericmodel_content_types():
    """
    Retrieve all models which inherit from GenericModel class
    and return their ContentType.
    """
    genericmodels = list(
        filter(
            lambda content_type: is_genericmodel(content_type),
            ContentType.objects.all(),
        )
    )
    return genericmodels


@register.filter
def get_attribute(obj, attribute):
    return getattr(obj, attribute, None)


@register.filter
def content_type_count(content_type):
    """
    Return the number of objects having a specific content type
    """
    return content_type.model_class().objects.count()


@register.simple_tag
def template_list(obj, suffix):
    return template_names_via_mro(type(obj), suffix)



# Generated by Django 4.2.11 on 2024-03-27 14:17

import django.db.models.deletion
import simple_history.models
from django.conf import settings
from django.db import migrations, models

import apis_core.apis_relations.models
import apis_core.generic.abc


class Migration(migrations.Migration):
    dependencies = [
        ("apis_metainfo", "0011_alter_rootobject_deprecated_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("apis_relations", "0005_alter_property_obj_class_alter_property_subj_class"),
    ]

    operations = [
        migrations.CreateModel(
            name="VersionTempTriple",
            fields=[
                (
                    "triple_ptr",
                    models.ForeignKey(
                        auto_created=True,
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        parent_link=True,
                        related_name="+",
                        to="apis_relations.triple",
                    ),
                ),
                (
                    "id",
                    models.IntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                (
                    "version_tag",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                (
                    "review",
                    models.BooleanField(
                        default=False,
                        help_text="Should be set to True, if the data record holds up quality standards.",
                    ),
                ),
                ("start_date", models.DateField(blank=True, null=True)),
                ("start_start_date", models.DateField(blank=True, null=True)),
                ("start_end_date", models.DateField(blank=True, null=True)),
                ("end_date", models.DateField(blank=True, null=True)),
                ("end_start_date", models.DateField(blank=True, null=True)),
                ("end_end_date", models.DateField(blank=True, null=True)),
                (
                    "start_date_written",
                    models.CharField(
                        blank=True, max_length=255, null=True, verbose_name="Start"
                    ),
                ),
                (
                    "end_date_written",
                    models.CharField(
                        blank=True, max_length=255, null=True, verbose_name="End"
                    ),
                ),
                ("status", models.CharField(max_length=100)),
                ("references", models.TextField(blank=True, null=True)),
                ("notes", models.TextField(blank=True, null=True)),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "obj",
                    apis_core.apis_relations.models.InheritanceForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="apis_metainfo.rootobject",
                        verbose_name="Object",
                    ),
                ),
                (
                    "prop",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="apis_relations.property",
                        verbose_name="Property",
                    ),
                ),
                (
                    "subj",
                    apis_core.apis_relations.models.InheritanceForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="apis_metainfo.rootobject",
                        verbose_name="Subject",
                    ),
                ),
            ],
            options={
                "verbose_name": "Version",
                "verbose_name_plural": "Versions",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(
                simple_history.models.HistoricalChanges,
                models.Model,
                apis_core.generic.abc.GenericModel,
            ),
        ),
    ]



# SPDX-FileCopyrightText: 2023 Birger Schacht
# SPDX-License-Identifier: MIT

from pathlib import Path

from django.test import TestCase

from apis_core.apis_entities.abc import E21_Person, E53_Place, E74_Group
from apis_core.utils import rdf

# use `curl -H "Accept: application/rdf+xml" -L $URI` to fetch data

testdata = Path(__file__).parent / "testdata"


class Place(E53_Place):
    class Meta:
        app_label = "test"


class Person(E21_Person):
    class Meta:
        app_label = "test"


class Institution(E74_Group):
    class Meta:
        app_label = "test"


class RdfTest(TestCase):
    def test_get_definition_from_dict_place_from_geonames(self):
        achensee = {
            "latitude": ["47.5"],
            "longitude": ["11.7"],
            "label": ["Achensee"],
        }
        # https://www.geonames.org/2783029/achensee.html
        uri = str(testdata / "achensee.rdf")

        place = Place()
        defintion, attributes = rdf.get_definition_and_attributes_from_uri(uri, place)
        self.assertEqual(achensee, attributes)

    def test_get_definition_from_dict_place_from_dnb(self):
        wien = {
            "label": ["Wien"],
            "latitude": ["048.208199"],
            "longitude": ["016.371690"],
        }
        # https://d-nb.info/gnd/4066009-6
        uri = str(testdata / "wien.rdf")

        place = Place()
        defintion, attributes = rdf.get_definition_and_attributes_from_uri(uri, place)
        self.assertEqual(wien, attributes)

    def test_get_definition_from_dict_person_from_dnb(self):
        pierre = {
            "forename": ["Pierre"],
            "surname": ["Ramus"],
            "alternative_names": [
                "Ramus, Pʹer",
                "Großmann, Rudolf",
                "Grossmann, Rudolf",
                "Grossman, Rudolf",
                "Grossman, Rodolphe",
                "Grossmann, Rodolphe",
                "Libertarian, ...",
            ],
            "date_of_birth": ["1882-04-15"],
            "date_of_death": ["1942"],
        }
        # https://d-nb.info/gnd/118833197
        uri = str(testdata / "ramus.rdf")

        person = Person()
        defintion, attributes = rdf.get_definition_and_attributes_from_uri(uri, person)
        self.assertEqual(pierre, dict(attributes))

    def test_get_definition_from_dict_institution_from_dnb(self):
        pierre_ges = {
            "label": ["Pierre-Ramus-Gesellschaft"],
        }
        # https://d-nb.info/gnd/415006-5
        uri = str(testdata / "ramus_gesellschaft.rdf")

        institution = Institution()
        defintion, attributes = rdf.get_definition_and_attributes_from_uri(
            uri, institution
        )
        self.assertEqual(pierre_ges, dict(attributes))

    def test_get_definition_from_dict_institution_from_dnb2(self):
        pierre_ges = {
            "label": ["Akademie der Wissenschaften in Wien"],
        }
        # https://d-nb.info/gnd/35077-1
        uri = str(testdata / "oeaw.rdf")

        institution = Institution()
        defintion, attributes = rdf.get_definition_and_attributes_from_uri(
            uri, institution
        )
        self.assertEqual(pierre_ges, dict(attributes))



from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django_tables2 import RequestConfig

from apis_core.utils.settings import get_entity_settings_by_modelname


def triple_sidebar_history(obj: object, request, detail=True):
    side_bar = []

    triples_related_all = obj.get_triples_for_version()
    if obj.version_tag is not None:
        triples_related_all = triples_related_all.filter(
            Q(version_tag=obj.version_tag)
            | Q(version_tag__contains=f"{obj.version_tag},")
        )
    content_type = ContentType.objects.get_for_model(obj.instance_type)
    if "apis_core.apis_relations" in settings.INSTALLED_APPS:
        from apis_core.apis_relations.tables import get_generic_triple_table
        from apis_core.apis_relations.utils import (
            get_content_types_with_allowed_relation_from,
        )

        for other_content_type in get_content_types_with_allowed_relation_from(
            content_type
        ):
            triples_related_by_entity = triples_related_all.filter(
                (
                    Q(subj__self_contenttype=other_content_type)
                    & Q(obj__pk=obj.instance.pk)
                )
                | (
                    Q(obj__self_contenttype=other_content_type)
                    & Q(subj__pk=obj.instance.pk)
                )
            )

            table_class = get_generic_triple_table(
                other_entity_class_name=other_content_type.model,
                entity_pk_self=obj.instance.pk,
                detail=detail,
            )

            prefix = f"{other_content_type.model}"
            title_card = other_content_type.name
            tb_object = table_class(data=triples_related_by_entity, prefix=prefix)
            tb_object_open = request.GET.get(prefix + "page", None)
            entity_settings = get_entity_settings_by_modelname(content_type.model)
            per_page = entity_settings.get("relations_per_page", 10)
            RequestConfig(request, paginate={"per_page": per_page}).configure(tb_object)
            tab_id = f"triple_form_{content_type.model}_to_{other_content_type.model}"
            side_bar.append(
                (
                    title_card,
                    tb_object,
                    tab_id,
                    tb_object_open,
                )
            )
    return side_bar



import logging

from django.contrib.contenttypes.models import ContentType
from django.dispatch import receiver

from apis_core.generic.signals import post_duplicate, post_merge_with
from apis_core.relations.models import Relation

logger = logging.getLogger(__name__)


@receiver(post_duplicate)
def copy_relations(sender, instance, duplicate, **kwargs):
    logger.info(f"Copying relations from {instance!r} to {duplicate!r}")
    content_type = ContentType.objects.get_for_model(instance)
    subj_rels = Relation.objects.filter(
        subj_content_type=content_type, subj_object_id=instance.id
    ).select_subclasses()
    obj_rels = Relation.objects.filter(
        obj_content_type=content_type, obj_object_id=instance.id
    ).select_subclasses()
    for rel in subj_rels:
        rel.pk = None
        rel.id = None
        rel.subj_object_id = duplicate.id
        rel.save()
    for rel in obj_rels:
        rel.pk = None
        rel.id = None
        rel.obj_object_id = duplicate.id
        rel.save()


@receiver(post_merge_with)
def merge_relations(sender, instance, entities, **kwargs):
    for ent in entities:
        logger.info(f"Merging relations from {ent!r} into {instance!r}")
        content_type = ContentType.objects.get_for_model(ent)
        Relation.objects.filter(
            subj_content_type=content_type, subj_object_id=ent.id
        ).update(subj_object_id=instance.id)
        Relation.objects.filter(
            obj_content_type=content_type, obj_object_id=ent.id
        ).update(obj_object_id=instance.id)



# Generated by Django 3.1.14 on 2022-02-01 12:41

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("apis_metainfo", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="VocabNames",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="VocabsBaseClass",
            fields=[
                (
                    "rootobject_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="apis_metainfo.rootobject",
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, help_text="Brief description of the used term."
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("rej", "rejected"),
                            ("ac", "accepted"),
                            ("can", "candidate"),
                            ("del", "deleted"),
                        ],
                        default="can",
                        max_length=4,
                    ),
                ),
                (
                    "parent_class",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="apis_vocabularies.vocabsbaseclass",
                    ),
                ),
                (
                    "userAdded",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "vocab_name",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="apis_vocabularies.vocabnames",
                    ),
                ),
            ],
            bases=("apis_metainfo.rootobject",),
        ),
        migrations.CreateModel(
            name="CollectionType",
            fields=[
                (
                    "vocabsbaseclass_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="apis_vocabularies.vocabsbaseclass",
                    ),
                ),
            ],
            bases=("apis_vocabularies.vocabsbaseclass",),
        ),
        migrations.CreateModel(
            name="LabelType",
            fields=[
                (
                    "vocabsbaseclass_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="apis_vocabularies.vocabsbaseclass",
                    ),
                ),
            ],
            bases=("apis_vocabularies.vocabsbaseclass",),
        ),
        migrations.CreateModel(
            name="VocabsUri",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uri", models.URLField()),
                ("domain", models.CharField(blank=True, max_length=255)),
                ("rdf_link", models.URLField(blank=True)),
                ("loaded", models.BooleanField(default=False)),
                ("loaded_time", models.DateTimeField(blank=True, null=True)),
                (
                    "vocab",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="apis_vocabularies.vocabsbaseclass",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="TextType",
            fields=[
                (
                    "vocabsbaseclass_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="apis_vocabularies.vocabsbaseclass",
                    ),
                ),
                ("entity", models.CharField(max_length=255)),
                (
                    "lang",
                    models.CharField(
                        blank=True,
                        default="deu",
                        help_text="The ISO 639-3 (or 2) code for the label's language.",
                        max_length=3,
                        null=True,
                        verbose_name="ISO Code",
                    ),
                ),
            ],
            bases=("apis_vocabularies.vocabsbaseclass",),
        ),
    ]



# Generated by Django 4.2.9 on 2024-01-26 07:57

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("collections", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="skoscollection",
            options={"ordering": ["name"]},
        ),
    ]



# Generated by Django 4.2.8 on 2023-12-22 07:56

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_vocabularies", "0003_delete_vocabnames_delete_vocabsuri"),
    ]

    operations = [
        migrations.DeleteModel(
            name="CollectionType",
        ),
    ]



from datetime import datetime, timedelta

from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from apis_core.apis_relations.models import Property, TempTriple
from sample_project.models import Person, Place, Profession


class SimpleHistoryTestCase(TestCase):
    """Test of the simple_history package using the demo project"""

    def setUp(self):
        self.Person = Person
        self.Place = Place
        self.Profession = Profession

        prop = Property.objects.create(
            name_forward="geboren in", name_reverse="Geburtsort von"
        )
        prop.subj_class.add(ContentType.objects.get(model="person"))
        prop.obj_class.add(ContentType.objects.get(model="place"))
        Person.objects.create(forename="John", surname="Doe")
        Place.objects.create(
            label="Steyr", _history_date=datetime.now() - timedelta(hours=0, minutes=50)
        )

    def test_history(self):
        """Tests the simple version of attributes changed on a models instance."""
        pers = self.Person.objects.get(forename="John")
        pers.forename = "Jane"
        pers.save()
        self.assertEqual(pers.forename, "Jane")
        pers_history = pers.history.all()
        assert len(pers_history) == 2
        self.assertEqual("Jane", pers.history.most_recent().forename)
        self.assertEqual("John", pers.history.earliest().forename)

    def test_history_through_triples(self):
        """Tests the newly introduced function for retrieving triples for a specific version of a model instance."""
        # first entity
        pers = self.Person.objects.get(forename="John")
        # second entity
        place = self.Place.objects.first()
        # create a triple
        tt1 = TempTriple.objects.create(
            subj=pers, prop=Property.objects.first(), obj=place
        )
        # test that we get one triple version for both entities
        self.assertEqual(len(pers.history.earliest().get_triples_for_version()), 1)
        self.assertEqual(len(place.history.earliest().get_triples_for_version()), 1)
        # change the triple and test again
        tt1.start_date_written = "2020-01-01"
        tt1.save()
        triples = pers.history.earliest().get_triples_for_version()
        self.assertEqual(len(triples), 1)
        self.assertEqual(triples[0].start_date_written, "2020-01-01")
        # change the place and test again
        place2 = self.Place.objects.create(label="testplace")
        tt1.obj = place2
        tt1.save()
        triples = place.history.earliest().get_triples_for_version()
        self.assertEqual(len(triples), 1)
        self.assertEqual(triples[0].obj, place)
        triples = place2.history.earliest().get_triples_for_version()
        self.assertEqual(len(triples), 1)
        self.assertEqual(triples[0].obj, place2)
        # given that the function should return the latest version of a triple, we should get the new place
        triples = pers.history.earliest().get_triples_for_version()
        self.assertEqual(len(triples), 1)
        self.assertEqual(triples[0].obj, place2)
        # test retrieving triples for a specific datetime
        triples = pers.history.earliest().get_triples_for_version(
            history_date=datetime.now() - timedelta(hours=0, minutes=40)
        )
        self.assertEqual(len(triples), 0)

    def test_history_tag(self):
        """Tests the version tag function."""
        pers = self.Person.objects.get(forename="John")
        pers.history.latest().set_version_tag("test_tag")
        self.assertEqual(pers.history.latest().version_tag, "test_tag")
        pers_history = pers.history.all()
        self.assertEqual(pers_history.count(), 1)
        self.assertEqual(pers_history[0].version_tag, "test_tag")
        # test with TemTriple
        pers2 = self.Person.objects.create(forename="Jane", surname="Doe")
        place = self.Place.objects.first()
        TempTriple.objects.create(subj=pers2, prop=Property.objects.first(), obj=place)
        pers2.history.latest().set_version_tag("test_tag")
        self.assertEqual(pers2.history.latest().version_tag, "test_tag")
        triples = pers2.history.earliest().get_triples_for_version()
        self.assertEqual(triples[0].version_tag, "test_tag")

    def test_history_delete_entry(self):
        """Tests the deletion of an entry."""
        pers = self.Person.objects.get(forename="John")
        pers.delete()
        assert len(self.Person.history.all()) == 2

    def test_history_merge(self):
        """Tests the merge function of the Place model. This is still expected to fail."""
        pl1 = self.Place.objects.first()
        pl2 = self.Place.objects.create(
            label="Test", _history_date=datetime.now() - timedelta(hours=0, minutes=10)
        )
        pl1.merge_with([pl2])
        print("save()")

    def test_m2m_save(self):
        """Test if m2m profession is saved correctly."""
        pers = self.Person.objects.create(
            forename="John",
            _history_date=datetime.now() - timedelta(hours=0, minutes=10),
        )
        self.assertEqual(pers.forename, "John")
        pers_history = pers.history.all()
        assert len(pers_history) == 1
        prof = self.Profession.objects.create(name="Test")
        pers.profession.add(prof)
        pers.forename = "Jane"
        pers.save()
        assert len(pers.profession.all()) == 1
        assert len(pers.history.latest().profession.all()) == 1
        assert len(pers.history.earliest().profession.all()) == 0

    def test_history_date(self):
        """Test that history is set correctly when not manually set."""
        pers = self.Person.objects.all().first()
        pers.forename = "Jane"
        pers.save()
        assert pers.history.earliest().history_date < pers.history.latest().history_date



import json
import logging
import urllib
from functools import cache

from AcdhArcheAssets.uri_norm_rules import get_normalized_uri
from django.core.exceptions import ImproperlyConfigured

from apis_core.utils.helpers import flatten_if_single
from apis_core.utils.rdf import get_definition_and_attributes_from_uri

logger = logging.getLogger(__name__)


class GenericModelImporter:
    """
    A generic importer class
    It provides the standard methods for importing data from
    an URI and creating a model instance of it.
    By default it fetches a resource, first tries to parse it using
    our rdf parser, if that fails tries to parse it using json and
    then extracts the fields whose keys match the model field names.
    Projects can inherit from this class and override the default
    methods or simple write their own from scratch.
    """

    model = None
    import_uri = None

    def __init__(self, uri, model):
        self.model = model
        self.import_uri = self.clean_uri(uri)

    @property
    def get_uri(self):
        return self.import_uri

    def clean_uri(self, uri):
        return get_normalized_uri(uri)

    @cache
    def request(self, uri):
        # we first try to use the RDF parser
        try:
            defn, data = get_definition_and_attributes_from_uri(uri, self.model)
            return data
        except Exception as e:
            logger.debug(e)
        # if everything else fails, try parsing json
        # if even that does not help, return an empty dict
        try:
            return json.loads(urllib.request.urlopen(uri).read())
        except Exception as e:
            logger.debug(e)
        return {}

    def mangle_data(self, data):
        return data

    def get_data(self, drop_unknown_fields=True):
        """
        fetch the data using the `request` method and
        mangle the data using the `mangle_data` method.
        If the `drop_unknown_fields` argument is true,
        remove all fields from the data dict that do not
        have an equivalent field in the model.
        """
        data = self.request(self.import_uri)
        data = self.mangle_data(data)
        if drop_unknown_fields:
            # we are dropping all fields that are not part of the model
            modelfields = [field.name for field in self.model._meta.fields]
            data = {key: data[key] for key in data if key in modelfields}
        data = {key: flatten_if_single(value) for key, value in data.items()}
        if not data:
            raise ImproperlyConfigured(
                f"Could not import {self.import_uri}. Data fetched was: {data}"
            )
        return data

    def import_into_instance(self, instance, fields="__all__"):
        data = self.get_data()
        if fields == "__all__":
            fields = data.keys()
        for field in fields:
            if hasattr(instance, field) and field in data.keys():
                setattr(instance, field, data[field])
        instance.save()

    def create_instance(self):
        return self.model.objects.create(**self.get_data(drop_unknown_fields=True))



from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django_filters import CharFilter, ModelMultipleChoiceFilter

from apis_core.apis_relations.models import Property
from apis_core.generic.filtersets import GenericFilterSet, GenericFilterSetForm

PROPERTY_EXCLUDES = [
    "self_contenttype",
    "property_class_uri",
    "rootobject_ptr",
]


class PropertyFilterSetForm(GenericFilterSetForm):
    columns_exclude = PROPERTY_EXCLUDES


class PropertySubjObjFilter(ModelMultipleChoiceFilter):
    def get_queryset(self, request):
        return ContentType.objects.filter(
            pk__in=Property.objects.all().values(self.field_name)
        )


class PropertyFilterSet(GenericFilterSet):
    name_forward = CharFilter(lookup_expr="icontains")
    name_reverse = CharFilter(lookup_expr="icontains")
    subj_class = PropertySubjObjFilter()
    obj_class = PropertySubjObjFilter()

    class Meta:
        form = PropertyFilterSetForm
        exclude = PROPERTY_EXCLUDES


class TripleFilterSet(GenericFilterSet):
    subj = CharFilter(method="subj_icontains")
    obj = CharFilter(method="obj_icontains")
    subj_class = PropertySubjObjFilter(
        label="Subj class",
        method="class_in",
    )
    obj_class = PropertySubjObjFilter(
        label="Obj class",
        method="class_in",
    )

    def subj_icontains(self, queryset, name, value):
        return queryset.filter(subj__name__icontains=value)

    def obj_icontains(self, queryset, name, value):
        return queryset.filter(obj__name__icontains=value)

    def class_in(self, queryset, name, value):
        # value is the list of contenttypes
        if value and name:
            name, _ = name.split("_")
            return queryset.filter(Q(**{f"{name}__self_contenttype__in": value}))
        return queryset


class HistoricalTripleFilterSet(TripleFilterSet):
    pass


class HistoricalTempTripleFilterSet(TripleFilterSet):
    pass



import functools
import logging

from django.contrib.auth import get_permission_codename
from django.db.models import CharField, Model, Q, TextField
from django.utils import module_loading

logger = logging.getLogger(__name__)


def default_search_fields(model, field_names=None):
    """
    Retrieve the default model fields to use for a search operation
    By default those are all the CharFields and TextFields of a model.
    It is also possible to define those fields on the model using the
    `_default_search_fields` attribute.
    The method also takes a `field_names` argument to override the list
    of fields.
    """
    default_types = (CharField, TextField)
    fields = [
        field for field in model._meta.get_fields() if isinstance(field, default_types)
    ]
    # check if the model has a `_default_search_fields`
    # list and use that as searchfields
    if isinstance(getattr(model, "_default_search_fields", None), list):
        fields = [
            model._meta.get_field(field) for field in model._default_search_fields
        ]
    # if `fields_to_search` is a list, use that
    if isinstance(field_names, list):
        fields = [model._meta.get_field(field) for field in field_names]
    return fields


def generate_search_filter(model, query, fields_to_search=None, prefix=""):
    """
    Generate a default search filter that searches for the `query`
    This helper can be used by autocomplete querysets if nothing
    fancier is needed.
    If the `prefix` is set, the field names will be prefixed with that string -
    this can be useful if you want to use the `generate_search_filter` in a
    `Q` combined query while searching over multiple models.
    """
    query = query.split()

    _fields_to_search = [
        field.name for field in default_search_fields(model, fields_to_search)
    ]

    q = Q()

    for token in query:
        q &= functools.reduce(
            lambda acc, field_name: acc
            | Q(**{f"{prefix}{field_name}__icontains": token}),
            _fields_to_search,
            Q(),
        )
    return q


@functools.lru_cache
def mro_paths(model):
    """
    Create a list of MRO classes for a Django model
    """
    paths = []
    for cls in filter(lambda x: x not in Model.mro(), model.mro()):
        paths.append(cls.__module__.split(".")[:-1] + [cls.__name__])
    return paths


@functools.lru_cache
def template_names_via_mro(model, suffix=""):
    """
    Use the MRO to generate a list of template names for a model
    """
    mro_prefix_list = ["/".join(prefix) for prefix in mro_paths(model)]
    return [f"{prefix.lower()}{suffix}" for prefix in mro_prefix_list]


@functools.lru_cache
def permission_fullname(action: str, model: object) -> str:
    permission_codename = get_permission_codename(action, model._meta)
    return f"{model._meta.app_label}.{permission_codename}"


@functools.lru_cache
def module_paths(model, path: str = "", suffix: str = "") -> list:
    paths = list(map(lambda x: x[:-1] + [path] + x[-1:], mro_paths(model)))
    classes = tuple(".".join(prefix) + suffix for prefix in paths)
    return classes


@functools.lru_cache
def makeclassprefix(string: str) -> str:
    string = "".join([c if c.isidentifier() else " " for c in string])
    string = "".join([word.strip().capitalize() for word in string.split(" ")])
    return string


@functools.lru_cache
def import_string(dotted_path):
    try:
        return module_loading.import_string(dotted_path)
    except (ModuleNotFoundError, ImportError):
        return False


@functools.lru_cache
def first_member_match(dotted_path_list: tuple[str], fallback=None) -> object:
    logger.debug("Looking for matching class in %s", dotted_path_list)
    pathgen = map(import_string, dotted_path_list)
    result = next(filter(bool, pathgen), None)
    if result:
        logger.debug("Found matching attribute/class in %s", result)
    else:
        logger.debug("Found nothing, returning fallback: %s", fallback)
    return result or fallback



# Generated by Django 5.1 on 2024-09-18 06:35

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_relations", "0006_versiontemptriple"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="versiontemptriple",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical temp triple",
                "verbose_name_plural": "historical temp triples",
            },
        ),
    ]



# Generated by Django 4.2.10 on 2024-03-15 10:02

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_metainfo", "0011_alter_rootobject_deprecated_name"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="rootobject",
            name="deprecated_name",
        ),
    ]



import logging

import django_filters
from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django_filters.filterset import FilterSet

from apis_core.generic.forms.fields import IncludeExcludeField

from .forms import GenericFilterSetForm

logger = logging.getLogger(__name__)


class CollectionsFilter(django_filters.filters.ModelMultipleChoiceFilter):
    """
    A simple filter for connections to collections. It provides an `include`/`exclude`
    selector, which allows to define what the filter should do.
    """

    @property
    def field(self):
        return IncludeExcludeField(super().field, required=self.extra["required"])

    def filter(self, queryset, value):
        if value:
            value, include_exclude = value
            content_type = ContentType.objects.get_for_model(queryset.model)
            try:
                skoscollectioncontentobject = apps.get_model(
                    "collections.SkosCollectionContentObject"
                )
                scco = skoscollectioncontentobject.objects.filter(
                    content_type=content_type, collection__in=value
                ).values("object_id")
                match include_exclude:
                    case "include":
                        return queryset.filter(id__in=scco)
                    case "exclude":
                        return queryset.exclude(id__in=scco)
            except LookupError as e:
                logger.debug("Not filtering for collections: %s", e)
        return queryset


class GenericFilterSet(FilterSet):
    """
    Our GenericFilterSet sets the default `form` to be our
    GenericFilterSetForm, which is set up to ignore the `columns` field
    of the form.
    """

    class Meta:
        form = GenericFilterSetForm

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            skoscollection = apps.get_model("collections.SkosCollection")
            if skoscollection.objects.exists():
                self.filters["collections"] = CollectionsFilter(
                    queryset=skoscollection.objects.all(),
                )
        except LookupError as e:
            logger.debug("Not adding collections filter to form: %s", e)



from django.contrib.contenttypes.models import ContentType
from django.shortcuts import redirect
from django.urls import reverse
from django.utils import timezone
from django.views.generic.detail import DetailView
from django_tables2 import SingleTableMixin
from django_tables2.tables import table_factory

from apis_core.generic.helpers import first_member_match, module_paths
from apis_core.generic.views import GenericModelMixin

from .tables import HistoryGenericTable


class HistoryView(GenericModelMixin, SingleTableMixin, DetailView):
    template_name = "history/history.html"

    def get_table_class(self):
        table_modules = module_paths(self.model, path="tables", suffix="HistoryTable")
        table_class = first_member_match(table_modules, HistoryGenericTable)
        return table_factory(self.model, table_class)

    def get_table_data(self):
        return self.get_object().get_history_data()


def create_new_version(request, contenttype, pk):
    """Gets the version of the history instance and creates a new version."""
    model = contenttype.model_class()
    instance = model.objects.get(id=pk)
    history_latest = instance.history.latest()
    latest_version_list = [
        int(x.replace("v", ""))
        for x in instance.history.filter(version_tag__isnull=False).values_list(
            "version_tag", flat=True
        )
    ]
    history_latest.history_id = None
    history_latest.history_date = timezone.now()
    history_latest.save()
    if latest_version_list:
        latest_version = max(latest_version_list)
    else:
        latest_version = 0
    history_latest.set_version_tag(f"v{latest_version + 1}")
    return redirect(
        reverse(
            "apis_core:generic:detail",
            args=[
                ContentType.objects.get_for_model(history_latest.__class__),
                history_latest.history_id,
            ],
        )
    )



# Generated by Django 5.1.1 on 2024-11-29 08:31

import django.db.models.deletion
from django.db import migrations, models


def copy_root_object(apps, schema_editor):
    Uri = apps.get_model("apis_metainfo", "Uri")
    RootObject = apps.get_model("apis_metainfo", "RootObject")
    for uri in Uri.objects.all():
        if uri.root_object_id:
            root_object = RootObject.objects.get(pk=uri.root_object_id)
            uri.content_type_id = root_object.self_contenttype_id
            uri.object_id = root_object.id
            uri.save()


def reverse_copy_root_object(apps, schema_editor):
    Uri = apps.get_model("apis_metainfo", "Uri")
    for uri in Uri.objects.all():
        if uri.content_type_id and uri.object_id:
            uri.root_object_id = uri.object_id
            uri.save()


class Migration(migrations.Migration):
    dependencies = [
        ("apis_metainfo", "0014_remove_uri_domain_remove_uri_loaded_and_more"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.AddField(
            model_name="uri",
            name="content_type",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="contenttypes.contenttype",
            ),
        ),
        migrations.AddField(
            model_name="uri",
            name="object_id",
            field=models.PositiveIntegerField(null=True),
        ),
        migrations.RunPython(copy_root_object, reverse_copy_root_object),
        migrations.RemoveField(model_name="uri", name="root_object"),
    ]



import math
import re
from datetime import datetime, timedelta


def parse_date(date_string: str) -> (datetime, datetime, datetime):
    """
    function to parse a string date field of an entity

    :param date_string : str :
        the field value passed by a user
    :return date_single : datetime :
        single date which represents either the precise date given by user or median in between a range.
    :return date_ab : datetime :
        starting date of a range if user passed a range value either implicit or explicit.
    :return date_bis : datetime :
        ending date of a range if user passed a range value either implicit or explicit.
    """

    def parse_date_range_individual(date, ab=False, bis=False):
        """
        As a sub function to parse_date, this function parse_date_individual handles a very single date since
        in a text field a user can pass multiple dates.

        :param date : str :
            recognized sub string which potentially is a date (in julian calendar format)
        :param ab : boolean : optional
            indicates if a single date shall be intepreted as a starting date of a range
        :param bis : boolean : optional
            indicates if a single date shall be intepreted as an ending date of a range
        :return tuple (datetime, datetime) :
            two datetime objects representing the dates.
            Two indicate that an implicit single date range was given (e.g. a year without months or days).
            Has to be further processed then since it can be either a starting or ending date range.
        or
        :return datetime :
            One datetime object representing the date.
            if a single date was given.
        """

        def get_last_day_of_month(month, year):
            """
            Helper function to return the last day of a given month and year (respecting leap years)

            :param month : int
            :param year : int
            :return day : int
            """

            if month in [1, 3, 5, 7, 8, 10, 12]:
                # 31 day months
                return 31
            elif month in [4, 6, 9, 11]:
                # 30 day months
                return 30
            elif month == 2:
                # special case february, differentiate leap years with respect to gregorian leap rules
                if year % 4 == 0:
                    if year % 100 == 0:
                        if year % 400 == 0:
                            # divisible by 4, by 100, by 400
                            # thus is leap year
                            return 29
                        else:
                            # divisible by 4, by 100, not by 400
                            # thus is not leap yar
                            return 28
                    else:
                        # divisible by 4, not by 100, if by 400 doesn't matter
                        # thus is leap year
                        return 29
                else:
                    # not divisible by 4, if by 100 or by 400 doesn't matter
                    return 28
            else:
                # no valid month
                raise ValueError("Month " + str(month) + " does not exist.")

        # replace all kinds of delimiters
        date = (
            date.replace(" ", "").replace("-", ".").replace("/", ".").replace("\\", ".")
        )
        # parse into variables for use later
        year = None
        month = None
        day = None
        # check for all kind of Y-M-D combinations
        if re.match(r"\d{3,4}$", date):
            # year
            year = int(date)
        elif re.match(r"\d{1,2}\.\d{3,4}$", date):
            # month - year
            tmp = re.split(r"\.", date)
            month = int(tmp[0])
            year = int(tmp[1])
        elif re.match(r"\d{1,2}\.\d{1,2}\.\d{3,4}$", date):
            # day - month - year
            tmp = re.split(r"\.", date)
            day = int(tmp[0])
            month = int(tmp[1])
            year = int(tmp[2])
        elif re.match(r"\d{3,4}\.\d{1,2}\.?$", date):
            # year - month
            tmp = re.split(r"\.", date)
            year = int(tmp[0])
            month = int(tmp[1])
        elif re.match(r"\d{3,4}\.\d{1,2}\.\d{1,2}\.?$", date):
            # year - month - day
            tmp = re.split(r"\.", date)
            year = int(tmp[0])
            month = int(tmp[1])
            day = int(tmp[2])
        else:
            # No sensical interpretation found
            raise ValueError("Could not interpret date.")
        if (ab and bis) or year is None:
            # both ab and bis in one single date are not valid, neither is the absence of a year.
            raise ValueError("Could not interpret date.")
        elif not ab and not bis and (month is None or day is None):
            # if both ab and bis are False and either month or day is empty, then it was given
            # an implicit date range (range of all months if given a year or all days if given a month)
            # construct implicit month range
            if month is None:
                month_ab = 1
                month_bis = 12
            else:
                month_ab = month
                month_bis = month
            # construct implicit day range
            if day is None:
                day_ab = 1
                day_bis = get_last_day_of_month(month_bis, year)
            else:
                day_ab = day
                day_bis = day

            # return a tuple from a single date (which the calling function has to further process)
            return (
                datetime(year=year, month=month_ab, day=day_ab),
                datetime(year=year, month=month_bis, day=day_bis),
            )
        else:
            # Either ab or bis is True. Then use the respective beginning or end of range and construct a precise date
            # Or both ab and bis are False. Then construct a precise date from parsed values
            # construct implicit month range if month is None
            if month is None:
                if ab and not bis:
                    # is a starting date, thus take first month of year
                    month = 1
                elif not ab and bis:
                    # is an ending date, thus take last month of year
                    month = 12
            # construct implicit day range if day is None
            if day is None:
                if ab and not bis:
                    # is a starting date, thus take first day of month
                    day = 1
                elif not ab and bis:
                    # is an ending date, thus take last month of year
                    day = get_last_day_of_month(month=month, year=year)

            return datetime(year=year, month=month, day=day)

    try:
        # return variables
        date_single = None
        date_ab = None
        date_bis = None
        # split for angle brackets, check if explicit iso date is contained within them
        date_split_angle = re.split(r"(<.*?>)", date_string)
        if len(date_split_angle) > 1:
            # date string contains angle brackets. Parse them, ignore the rest
            def parse_iso_date(date_string):
                date_string_split = date_string.split("-")
                try:
                    return datetime(
                        year=int(date_string_split[0]),
                        month=int(date_string_split[1]),
                        day=int(date_string_split[2]),
                    )
                except Exception:
                    raise ValueError("Invalid iso date: ", date_string)

            if len(date_split_angle) > 3:
                # invalid case
                raise ValueError("Too many angle brackets.")
            elif len(date_split_angle) == 3:
                # the right amount of substrings, indicating exactly one pair of angle brackets.
                # Parse the iso date in between
                # remove angle brackets and split by commas
                dates_iso = date_split_angle[1][1:-1]
                # check for commas, which would indicate that either one iso date or three are being input
                dates_iso = dates_iso.split(",")
                if len(dates_iso) != 1 and len(dates_iso) != 3:
                    # only either one iso date or three are allowed
                    raise ValueError(
                        "Incorrect number of dates given. Within angle brackets only one or three (separated by commas) are allowed."
                    )
                elif len(dates_iso) == 3:
                    # three iso dates indicate further start and end dates
                    # parse start date
                    date_ab_string = dates_iso[1].strip()
                    if date_ab_string != "":
                        date_ab = parse_iso_date(date_ab_string)
                    # parse end date
                    date_bis_string = dates_iso[2].strip()
                    if date_bis_string != "":
                        date_bis = parse_iso_date(date_bis_string)
                # parse single date
                date_single_string = dates_iso[0].strip()
                if date_single_string != "":
                    date_single = parse_iso_date(date_single_string)
        else:
            # date string contains no angle brackets. Interpret the possible date formats
            date_string = date_string.lower()
            date_string = date_string.replace(" ", "")
            # helper variables for the following loop
            found_ab = False
            found_bis = False
            found_single = False
            # split by allowed keywords 'ab' and 'bis' and iterate over them
            date_split_ab_bis = re.split(r"(ab|bis)", date_string)
            for i, v in enumerate(date_split_ab_bis):
                if v == "ab":
                    # indicates that the next value must be a start date
                    if found_ab or found_single:
                        # if already found a ab_date or single date before then there is non-conformative redundancy
                        raise ValueError("Redundant dates found.")
                    found_ab = True
                    # parse the next value which must be a parsable date string
                    date_ab = parse_date_range_individual(
                        date_split_ab_bis[i + 1], ab=True
                    )
                elif v == "bis":
                    # indicates that the next value must be an end date
                    if found_bis or found_single:
                        # if already found a bis_date or single date before then there is non-conformative redundancy
                        raise ValueError("Redundant dates found.")
                    found_bis = True

                    # parse the next value which must be a parsable date string
                    date_bis = parse_date_range_individual(
                        date_split_ab_bis[i + 1], bis=True
                    )
                elif v != "" and not found_ab and not found_bis and not found_single:
                    # indicates that this value must be a date
                    found_single = True
                    # parse the this value which must be a parsable date string
                    date_single = parse_date_range_individual(v)
                    if type(date_single) is tuple:
                        #  if result of parse_date_range_individual is a tuple then the date was an implict range.
                        #  Then split it into start and end dates
                        date_ab = date_single[0]
                        date_bis = date_single[1]
            if date_ab and date_bis:
                # date is a range
                if date_ab > date_bis:
                    raise ValueError("'ab-date' must be before 'bis-date' in time")
                # calculate difference between start and end date of range,
                # and use it to calculate a single date for usage as median.
                days_delta_half = math.floor(
                    (date_bis - date_ab).days / 2,
                )
                date_single = date_ab + timedelta(days=days_delta_half)
            elif date_ab is not None and date_bis is None:
                # date is only the start of a range, save it also as the single date
                date_single = date_ab
            elif date_ab is None and date_bis is not None:
                # date is only the end of a range, save it also as the single date
                date_single = date_bis

    except Exception as e:
        print("Could not parse date: '", date_string, "' due to error: ", e)

    return date_single, date_ab, date_bis



from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from .models import RootObject, Uri


class ModelTestCase(TestCase):
    def setUp(cls):
        # Set up data for the whole TestCase
        user_type = ContentType.objects.get(app_label="auth", model="user")
        RootObject.objects.create(self_contenttype=user_type)

    def test_uri(self):
        ufoo = Uri.objects.create()
        self.assertEqual(str(ufoo), "None")



from rest_framework import viewsets

from apis_core.generic.schema import GenericAutoSchema

from .filterbackends import GenericFilterBackend
from .helpers import first_member_match, makeclassprefix, module_paths
from .serializers import GenericHyperlinkedModelSerializer, serializer_factory


class ModelViewSet(viewsets.ModelViewSet):
    """
    API ViewSet for a generic model.
    The queryset is overridden by the first match from
    the `first_member_match` helper.
    The serializer class is overridden by the first match from
    the `first_member_match` helper.
    """

    filter_backends = [GenericFilterBackend]
    schema = GenericAutoSchema()

    def dispatch(self, *args, **kwargs):
        self.model = kwargs.get("contenttype").model_class()
        return super().dispatch(*args, **kwargs)

    def get_queryset(self):
        queryset_methods = module_paths(
            self.model, path="querysets", suffix="ViewSetQueryset"
        )
        queryset = first_member_match(queryset_methods) or (lambda x: x)
        return queryset(self.model.objects.all())

    def get_serializer_class(self):
        renderer = getattr(getattr(self, "request", {}), "accepted_renderer", None)
        serializer_class_modules = module_paths(
            self.model, path="serializers", suffix="Serializer"
        )
        if renderer is not None:
            prefix = makeclassprefix(renderer.format)
            serializer_class_modules = (
                module_paths(
                    self.model, path="serializers", suffix=f"{prefix}Serializer"
                )
                + serializer_class_modules
            )

        serializer_class = first_member_match(
            serializer_class_modules,
            getattr(renderer, "serializer", GenericHyperlinkedModelSerializer),
        )
        return serializer_factory(self.model, serializer=serializer_class)



# SPDX-FileCopyrightText: 2023 Birger Schacht
# SPDX-License-Identifier: MIT

from datetime import datetime

from django.test import TestCase

from .DateParser import parse_date

help_text_default = "Dates are interpreted by defined rules. If this fails, an iso-date can be explicitly set with '&lt;YYYY-MM-DD&gt;'."


# helper function so we don't have to write as much
def fi(datestring):
    return datetime.fromisoformat(datestring)


# a dict of dates with the written date as key
# followed by a tuple of (single, start, end)
dates = {
    "vor dem 23.10.1449 <1449-10-23>": (
        fi("1449-10-23"),
        None,
        None,
    ),
    "1459-12": (
        fi("1459-12-16"),
        fi("1459-12-01"),
        fi("1459-12-31"),
    ),
    "1460-03-30": (fi("1460-03-30"), None, None),
    "": (
        None,
        None,
        None,
    ),
}


class DateParserTest(TestCase):
    def test_dates(self):
        for datestring, (expsingle, expstart, expend) in dates.items():
            single, start, end = parse_date(datestring)
            self.assertEqual(expsingle, single)
            self.assertEqual(expstart, start)
            self.assertEqual(expend, end)



# Generated by Django 3.1.14 on 2022-02-01 12:41

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Collection",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("description", models.TextField(blank=True)),
                ("published", models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name="RootObject",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255, verbose_name="Name")),
            ],
        ),
        migrations.CreateModel(
            name="Source",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("orig_filename", models.CharField(blank=True, max_length=255)),
                ("indexed", models.BooleanField(default=False)),
                ("pubinfo", models.CharField(blank=True, max_length=400)),
                ("author", models.CharField(blank=True, max_length=255)),
                ("orig_id", models.PositiveIntegerField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name="Text",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("text", models.TextField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name="UriCandidate",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uri", models.URLField()),
                ("confidence", models.FloatField(blank=True, null=True)),
                ("responsible", models.CharField(max_length=255)),
                (
                    "root_object",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="apis_metainfo.rootobject",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Uri",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "uri",
                    models.URLField(blank=True, max_length=255, null=True, unique=True),
                ),
                ("domain", models.CharField(blank=True, max_length=255)),
                ("rdf_link", models.URLField(blank=True)),
                ("loaded", models.BooleanField(default=False)),
                ("loaded_time", models.DateTimeField(blank=True, null=True)),
                (
                    "root_object",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="apis_metainfo.rootobject",
                    ),
                ),
            ],
        ),
    ]



import logging

from django.dispatch import receiver

from apis_core.apis_metainfo.models import RootObject
from apis_core.apis_relations.models import TempTriple
from apis_core.generic.signals import post_duplicate, post_merge_with

logger = logging.getLogger(__name__)


@receiver(post_duplicate)
def copy_relations(sender, instance, duplicate, **kwargs):
    logger.info(f"Copying relations from {instance} to {duplicate}")
    if isinstance(instance, RootObject):
        for rel in TempTriple.objects.filter(subj=instance):
            newrel = rel.duplicate()
            newrel.subj = duplicate
            newrel.save()
        for rel in TempTriple.objects.filter(obj=instance):
            newrel = rel.duplicate()
            newrel.obj = duplicate
            newrel.save()


@receiver(post_merge_with)
def merge_relations(sender, instance, entities, **kwargs):
    for ent in entities:
        logger.info(f"Merging relations from {ent!r} into {instance!r}")
        TempTriple.objects.filter(obj__id=ent.id).update(obj=instance)
        TempTriple.objects.filter(subj__id=ent.id).update(subj=instance)



# Generated by Django 4.2.10 on 2024-03-07 08:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("apis_metainfo", "0010_rename_name_rootobject_deprecated_name"),
    ]

    operations = [
        migrations.AlterField(
            model_name="rootobject",
            name="deprecated_name",
            field=models.CharField(blank=True, max_length=255, verbose_name="Name"),
        ),
    ]



from apis_ontology.models import construct_properties
from django.core.management.base import BaseCommand, CommandError

from apis_core.apis_relations.models import Property


class Command(BaseCommand):
    help = (
        "Create and label relationships between entities by running "
        "the construct_properties() function in your app's models.py file."
    )

    def handle(self, *args, **options):
        try:
            construct_properties()
            self.stdout.write(
                self.style.SUCCESS("Successfully ran construct_properties()")
            )
        except Exception as e:
            raise CommandError(
                f"construct_properties() raised the " f"following error(s):\n{e}"
            )

        try:
            props = Property.objects.all()
        except Property.DoesNotExist:
            raise CommandError("No entity relationships exist.")

        if len(props) > 0:
            # TODO would ideally verify existence/validity of relevant fields
            #  as relationship names alone are useless w/o subj/obj references;
            #  below msg. purposefully only mentions "labels".
            self.stdout.write("The following labels now exist:")
            for p in props:
                self.stdout.write(self.style.SUCCESS(f"{p.name_forward}"), ending="")
                self.stdout.write(f" ... {p.name_reverse}")
        else:
            # construct_properties() starts out by deleting all existing
            # Property objects
            self.stdout.write(
                self.style.ERROR("No entity relationships found nor created.")
            )



from django.contrib.contenttypes.models import ContentType
from django.http import Http404
from django.shortcuts import get_list_or_404
from django.urls import include, path, register_converter

from apis_core.apis_entities.api_views import GetEntityGeneric, ListEntityGeneric

# from .views import ReversionCompareView TODO: add again when import is fixed
from apis_core.apis_entities.models import AbstractEntity
from apis_core.apis_entities.views import (
    EntitiesAutocomplete,
    EntitiesDuplicate,
)

api_routes = [
    path("entities/", ListEntityGeneric.as_view()),
    path(
        "entity/<int:pk>/",
        GetEntityGeneric.as_view(),
        name="GetEntityGeneric",
    ),
]


class EntityToContenttypeConverter:
    """
    A converter that converts from a the name of an entity class
    (i.e. `person`) to the actual Django model class.
    """

    regex = r"\w+"

    def to_python(self, value):
        candiates = get_list_or_404(ContentType, model=value)
        candiates = list(
            filter(
                lambda ct: ct.model_class() is not None
                and issubclass(ct.model_class(), AbstractEntity),
                candiates,
            )
        )
        if len(candiates) > 1:
            raise Http404("Multiple entities match the <%s> identifier" % value)
        return candiates[0]

    def to_url(self, value):
        if isinstance(value, ContentType):
            return value.model
        if isinstance(value, str):
            return value


register_converter(EntityToContenttypeConverter, "entitytocontenttype")

app_name = "apis_entities"

entity_patterns = [
    path(
        "<int:pk>/duplicate/",
        EntitiesDuplicate.as_view(),
        name="generic_entities_duplicate_view",
    ),
]

urlpatterns = [
    path(
        "entity/<entitytocontenttype:contenttype>/",
        include(entity_patterns),
    ),
    path("autocomplete/", EntitiesAutocomplete.as_view(), name="autocomplete"),
]



from django.db import models

#########################
# Abstract base classes #
#########################


# These abstract base classes are named after
# CIDOC CRM entities, but we are *NOT*(!)
# trying to implement CIDOC CRM in Django.


class E21_Person(models.Model):
    forename = models.CharField(blank=True, default="", max_length=4096)
    surname = models.CharField(blank=True, default="", max_length=4096)
    gender = models.CharField(blank=True, default="", max_length=4096)
    date_of_birth = models.DateField(blank=True, null=True)
    date_of_death = models.DateField(blank=True, null=True)

    class Meta:
        abstract = True

    def __str__(self):
        return f"{self.forename} {self.surname}"


class E53_Place(models.Model):
    label = models.CharField(blank=True, default="", max_length=4096)
    latitude = models.FloatField(blank=True, null=True)
    longitude = models.FloatField(blank=True, null=True)

    class Meta:
        abstract = True

    def __str__(self):
        return self.label


class E74_Group(models.Model):
    label = models.CharField(blank=True, default="", max_length=4096)

    class Meta:
        abstract = True

    def __str__(self):
        return self.label



# Generated by Django 3.1.14 on 2023-01-23 17:15

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_relations", "0002_property_property_class_uri"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="property",
            options={"verbose_name_plural": "Properties"},
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from django.urls import resolve, reverse
from drf_spectacular.generators import EndpointEnumerator, SchemaGenerator

from apis_core.generic.abc import GenericModel

# Custom Schema Generator
#
# The CustomSchemaGenerator is meant as a drop in replacement for the
# "DEFAULT_GENERATOR_CLASS" of drf-spectacular. You can use it like this:
# SPECTACULAR_SETTINGS["DEFAULT_GENERATOR_CLASS"] = 'apis_core.generic.generators.CustomSchemaGenerator'
#
# Reasons:
# The first reason is, that we are using a custom converter
# (`apis_core.generic.urls.ContenttypeConverter`) to provide access to views
# and api views of different contenttypes. The default endpoint generator of
# does not know about this converter and therefore sees the endpoints using
# this converter as *one* endpoint with one parameter called `contenttype`.
# Thats not what we want, so we have to do our own enumeration - we iterate
# through all contenttypes that inherit from `GenericModel` and create
# schema endpoints for those programatically.
# The second reason is, that the autoapi schema generator of DRF Spectacular
# and our `apis_core.generic.api_views.ModelViewSet` don't work well together.
# Our ModelViewSet needs the dispatch method to set the model of the
# `ModelViewSet`, but this method is not being called by the generator while
# doing the inspection, which means the `ModelViewSet` does not know about the
# model it should work with and can not provide the correct serializer and filter
# classes. Therefore we create the callback for the endpoints by hand and set
# the model from there.


class CustomEndpointEnumerator(EndpointEnumerator):
    def _generate_content_type_endpoint(
        self, content_type: ContentType, method: str = "list"
    ):
        """Create a endpoint tuple, usable by the SchemaGenerator of DRF spectacular"""
        path = reverse("apis_core:generic:genericmodelapi-list", args=[content_type])
        cls = resolve(path).func.cls

        if method == "detail":
            path += "{id}/"
        regex = path
        # for now we only do "GET"
        httpmethod = "GET"
        # we have to add a attribute, so that the
        # `initkwargs` argument to the `as_view`
        # method can contain a `model` argument
        cls.model = None
        callback = cls.as_view({"get": method}, model=content_type.model_class())
        return (path, regex, httpmethod, callback)

    def get_api_endpoints(self, patterns=None, prefix=""):
        """
        Call the EndpointEnumerator's `get_api_endpoints` method to get all
        the automatically found endpoints, remove the ones that we want to override
        and the add our custom endpoints to this list.
        """
        api_endpoints = super().get_api_endpoints(patterns, prefix)
        api_endpoints = [
            endpoint
            for endpoint in api_endpoints
            if not endpoint[0].startswith("/apis/api/{contenttype}/")
        ]
        for content_type in ContentType.objects.all():
            if content_type.model_class() is not None and issubclass(
                content_type.model_class(), GenericModel
            ):
                api_endpoints.append(self._generate_content_type_endpoint(content_type))
                api_endpoints.append(
                    self._generate_content_type_endpoint(content_type, "detail")
                )
        return api_endpoints


class CustomSchemaGenerator(SchemaGenerator):
    endpoint_inspector_cls = CustomEndpointEnumerator



import logging

from rdflib import Graph
from rest_framework import renderers

logger = logging.getLogger(__name__)


class GenericRDFBaseRenderer(renderers.BaseRenderer):
    def render(self, data, accepted_media_type=None, renderer_context=None):
        g = Graph()
        for result in data.get("results", []):
            match result:
                case tuple(_, _, _):
                    g.add(result)
                case other:
                    logger.debug("Could not add %s to RDF graph: not a tuple", other)
        return g.serialize(format=self.media_type)


class GenericRDFXMLRenderer(GenericRDFBaseRenderer):
    media_type = "application/rdf+xml"
    format = "rdf+xml"


class GenericRDFTurtleRenderer(GenericRDFBaseRenderer):
    media_type = "text/turtle"
    format = "rdf+turtle"


class GenericRDFN3Renderer(GenericRDFBaseRenderer):
    media_type = "text/n3"
    format = "rdf+n3"



from AcdhArcheAssets.uri_norm_rules import get_normalized_uri
from django.core.management.base import BaseCommand
from django.db import transaction

from apis_core.apis_metainfo.models import Uri


class Command(BaseCommand):
    # Show this when the user types help
    help = "normalizes Uris"

    def handle(self, *args, **options):
        with transaction.atomic():
            for x in Uri.objects.all():
                old_uri = x.uri
                new_uri = get_normalized_uri(old_uri)
                if old_uri != new_uri:
                    x.uri = new_uri
                    x.save()
        return "all done"



# Generated by Django 4.2.8 on 2023-12-22 07:50

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_vocabularies", "0001_initial"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="texttype",
            name="vocabsbaseclass_ptr",
        ),
        migrations.RemoveField(
            model_name="vocabnames",
            name="name",
        ),
        migrations.RemoveField(
            model_name="vocabsbaseclass",
            name="description",
        ),
        migrations.RemoveField(
            model_name="vocabsbaseclass",
            name="parent_class",
        ),
        migrations.RemoveField(
            model_name="vocabsbaseclass",
            name="status",
        ),
        migrations.RemoveField(
            model_name="vocabsbaseclass",
            name="userAdded",
        ),
        migrations.RemoveField(
            model_name="vocabsbaseclass",
            name="vocab_name",
        ),
        migrations.RemoveField(
            model_name="vocabsuri",
            name="domain",
        ),
        migrations.RemoveField(
            model_name="vocabsuri",
            name="loaded",
        ),
        migrations.RemoveField(
            model_name="vocabsuri",
            name="loaded_time",
        ),
        migrations.RemoveField(
            model_name="vocabsuri",
            name="rdf_link",
        ),
        migrations.RemoveField(
            model_name="vocabsuri",
            name="uri",
        ),
        migrations.RemoveField(
            model_name="vocabsuri",
            name="vocab",
        ),
        migrations.DeleteModel(
            name="LabelType",
        ),
        migrations.DeleteModel(
            name="TextType",
        ),
    ]



import copy

from crispy_forms.helper import FormHelper
from dal import autocomplete
from django import forms
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.urls import reverse
from django_tables2 import RequestConfig

from apis_core.apis_entities.autocomplete3 import (
    PropertyAutocomplete,
)
from apis_core.apis_entities.fields import ListSelect2
from apis_core.apis_entities.utils import get_entity_classes
from apis_core.apis_metainfo.models import Uri
from apis_core.apis_relations.models import TempTriple
from apis_core.utils.settings import get_entity_settings_by_modelname

from .tables import get_generic_triple_table


class GenericTripleForm(forms.ModelForm):
    # TODO RDF : Add Notes and references

    class Meta:
        model = TempTriple
        fields = [
            "subj",
            "obj",
            "prop",
            "start_date_written",
            "end_date_written",
            "notes",
        ]
        widgets = {
            "subj": forms.HiddenInput(),
            "obj": forms.HiddenInput(),
            "prop": forms.HiddenInput(),
        }

    def __init__(self, entity_type_self_str, entity_type_other_str):
        super().__init__()

        self.helper = FormHelper()
        self.helper.form_tag = False

        attrs = {
            "data-placeholder": "Type to get suggestions",
            "data-minimum-input-length": getattr(settings, "APIS_MIN_CHAR", 3),
            "data-html": True,
            "style": "width: 100%",
        }
        help_text_other_entity = (
            "Search and select or use an URL from a reference resource"
        )
        attrs_target = copy.deepcopy(attrs)
        attrs_target["data-tags"] = "1"

        contenttypes = [
            ContentType.objects.get_for_model(model) for model in get_entity_classes()
        ]
        ct = next(
            filter(lambda x: x.model == entity_type_other_str.lower(), contenttypes)
        )
        url = reverse("apis_core:generic:autocomplete", args=[ct]) + "?create=True"

        self.fields["other_entity"] = autocomplete.Select2ListCreateChoiceField(
            label="entity",
            widget=ListSelect2(
                url=url,
                attrs=attrs_target,
            ),
            help_text=help_text_other_entity,
        )

        # This assert only serves as a linking for us devs, to make explicit what internal object the class
        # Select2ListCreateChoiceField object afterwards uses.
        assert PropertyAutocomplete

        self.fields["property"] = autocomplete.Select2ListCreateChoiceField(
            label="property",
            widget=ListSelect2(
                url=reverse(
                    "apis_core:apis_relations:generic_property_autocomplete",
                    kwargs={
                        "entity_self": entity_type_self_str,
                        "entity_other": entity_type_other_str,
                    },
                ),
                attrs=attrs_target,
            ),
        )
        self.helper.include_media = False

    def load_subj_obj_prop(
        self,
        entity_instance_self,
        entity_instance_other,
        property_instance,
        property_direction,
    ):
        # the more important function here when writing data from an user input via an ajax call into this form.
        # Because here the direction of the property is respected. Hence the subject and object position of the
        # triple and the property name_forward or name_reverse are loaded correctly here.

        if property_direction == PropertyAutocomplete.SELF_SUBJ_OTHER_OBJ_STR:
            triple_subj = entity_instance_self
            triple_obj = entity_instance_other
            property_direction_name = property_instance.name_forward

        elif property_direction == PropertyAutocomplete.SELF_OBJ_OTHER_SUBJ_STR:
            triple_subj = entity_instance_other
            triple_obj = entity_instance_self
            property_direction_name = property_instance.name_reverse

        else:
            raise Exception("No valid property direction given.")

        self.fields["subj"].initial = triple_subj
        self.fields["obj"].initial = triple_obj
        self.fields["prop"].initial = property_instance

        property_initial_value = (
            f"id:{property_instance.pk}__direction:{property_direction}",
            property_direction_name,
        )
        self.fields["property"].initial = property_initial_value
        self.fields["property"].choices = [property_initial_value]

        content_type = ContentType.objects.get_for_model(entity_instance_other)
        other_entity_initial_value = (
            str(
                Uri.objects.filter(
                    content_type=content_type, object_id=entity_instance_other.id
                ).first()
            ),
            f"<span ><small>db</small> {str(entity_instance_other)}</span>",
        )
        self.fields["other_entity"].initial = other_entity_initial_value
        self.fields["other_entity"].choices = [other_entity_initial_value]

    def load_remaining_data_from_triple(self, triple):
        # Most data is loaded via the set_subj_obj function.
        # Here, load the rest from a pre-existing triple.
        #
        # This function is both used in get_form_ajax and save_form_ajax,
        # hence it's not feasible to assume existing user input as done
        # in 'load_remaining_data_from_input'

        self.fields["start_date_written"].initial = triple.start_date_written
        self.fields["end_date_written"].initial = triple.end_date_written
        self.fields["notes"].initial = triple.notes
        self.instance = triple

    def load_remaining_data_from_input(
        self,
        start_date_written,
        end_date_written,
    ):
        # Most data is loaded via the set_subj_obj function.
        # Here, load the rest from the user input via the ajax post

        self.fields["start_date_written"].initial = start_date_written
        self.fields["end_date_written"].initial = end_date_written

    def save(self):
        if self.instance.pk is None:
            self.instance = TempTriple.objects.create(
                subj=self.fields["subj"].initial,
                obj=self.fields["obj"].initial,
                prop=self.fields["prop"].initial,
            )

        else:
            self.instance.subj = self.fields["subj"].initial
            self.instance.obj = self.fields["obj"].initial
            self.instance.prop = self.fields["prop"].initial

        self.instance.start_date_written = self.fields["start_date_written"].initial
        self.instance.end_date_written = self.fields["end_date_written"].initial
        self.instance.notes = self.fields["notes"].initial
        self.instance.save()

        return self.instance

    def get_text_id(self):
        """
        Function to retrieve the highlighted text.
        :return: ID of text that was highlighted
        """
        return self.cleaned_data["HL_text_id"][5:]

    def get_html_table(self, entity_instance_self, entity_instance_other, request):
        table_class = get_generic_triple_table(
            other_entity_class_name=entity_instance_other.__class__.__name__.lower(),
            entity_pk_self=entity_instance_self.pk,
            detail=False,
        )

        table_object = table_class(
            data=TempTriple.objects.filter(
                (
                    Q(subj__self_contenttype=entity_instance_other.self_contenttype)
                    & Q(obj=entity_instance_self)
                )
                | (
                    Q(obj__self_contenttype=entity_instance_other.self_contenttype)
                    & Q(subj=entity_instance_self)
                )
            ),
            prefix=entity_instance_other.__class__.__name__,
            request=request,
        )
        entity_settings = get_entity_settings_by_modelname(
            entity_instance_self.__class__.__name__
        )
        per_page = entity_settings.get("relations_per_page", 10)
        RequestConfig(request, paginate={"per_page": per_page}).configure(table_object)

        return table_object



# Generated by Django 4.2.10 on 2024-02-26 13:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("apis_relations", "0004_auto_20230124_1201"),
    ]

    operations = [
        migrations.AlterField(
            model_name="property",
            name="obj_class",
            field=models.ManyToManyField(
                related_name="property_set_obj", to="contenttypes.contenttype"
            ),
        ),
        migrations.AlterField(
            model_name="property",
            name="subj_class",
            field=models.ManyToManyField(
                related_name="property_set_subj", to="contenttypes.contenttype"
            ),
        ),
    ]



from django.db import models

from apis_core.utils import DateParser


# This is abstract class provides a Mixin for APIS models that need
# a temporal component. It provides commonly used felds for start
# and end dates, reps. start and end durations.
class LegacyDateMixin(models.Model):
    start_date = models.DateField(blank=True, null=True, editable=False)
    start_start_date = models.DateField(blank=True, null=True, editable=False)
    start_end_date = models.DateField(blank=True, null=True, editable=False)
    end_date = models.DateField(blank=True, null=True, editable=False)
    end_start_date = models.DateField(blank=True, null=True, editable=False)
    end_end_date = models.DateField(blank=True, null=True, editable=False)
    start_date_written = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name="Start",
    )
    end_date_written = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name="End",
    )

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        skip_date_parsing = getattr(self, "skip_date_parsing", False)
        if not skip_date_parsing:
            start = None
            start_start = None
            start_end = None
            end = None
            end_start = None
            end_end = None

            if self.start_date_written:
                start, start_start, start_end = DateParser.parse_date(
                    self.start_date_written
                )
                # DateParser returns datetime, but we want dates without time
                if start:
                    start = start.date()
                if start_start:
                    start_start = start_start.date()
                if start_end:
                    start_end = start_end.date()

            if self.end_date_written:
                end, end_start, end_end = DateParser.parse_date(self.end_date_written)
                # DateParser returns datetime, but we want dates without time
                if end:
                    end = end.date()
                if end_start:
                    end_start = end_start.date()
                if end_end:
                    end_end = end_end.date()

            self.start_date = start
            self.start_start_date = start_start
            self.start_end_date = start_end
            self.end_date = end
            self.end_start_date = end_start
            self.end_end_date = end_end

        super().save(*args, **kwargs)

        return self



import copy
import unicodedata

from crum import get_current_request
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Q
from django.db.models.fields.related_descriptors import ForwardManyToOneDescriptor
from django.db.models.signals import m2m_changed
from model_utils.managers import InheritanceManager

from apis_core.apis_metainfo.models import RootObject
from apis_core.generic import signals
from apis_core.generic.abc import GenericModel
from apis_core.history.models import VersionMixin
from apis_core.utils import DateParser


def find_if_user_accepted():
    request = get_current_request()
    if request is not None:
        if request.user.is_authenticated:
            return {}
        else:
            return {"published": True}
    else:
        return {}


class BaseRelationManager(models.Manager):
    def get_queryset(self):
        return RelationPublishedQueryset(self.model, using=self._db)

    def filter_ann_proj(self, request=None, ann_proj=1, include_all=True):
        return self.get_queryset().filter_ann_proj(
            request=request, ann_proj=ann_proj, include_all=include_all
        )

    def filter_for_user(self):
        if hasattr(settings, "APIS_SHOW_ONLY_PUBLISHED"):
            return self.get_queryset().filter_for_user()
        else:
            return self.get_queryset()


class Property(RootObject):
    class Meta:
        verbose_name_plural = "Properties"

    objects = BaseRelationManager()
    property_class_uri = models.CharField(
        max_length=255, verbose_name="Property Class URI", blank=True
    )
    # TODO RDF: Redundancy between name_forward and name, solve this.
    name_forward = models.CharField(
        max_length=255,
        verbose_name="Name forward",
        help_text='Inverse relation like: "is sub-class of" vs. "is super-class of".',
        blank=True,
    )
    # TODO RDF: Maybe rename name to name_subj_to_obj and name_reverse to name_obj_to_subj
    name_reverse = models.CharField(
        max_length=255,
        verbose_name="Name reverse",
        help_text='Inverse relation like: "is sub-class of" vs. "is super-class of".',
        blank=True,
    )
    subj_class = models.ManyToManyField(
        ContentType,
        related_name="property_set_subj",
    )
    obj_class = models.ManyToManyField(
        ContentType,
        related_name="property_set_obj",
    )

    def __str__(self):
        return self.name_forward

    def save(self, *args, **kwargs):
        if self.name_reverse == "":
            self.name_reverse = f"{self.name_forward} [INVERSE]"

        self.name_forward = unicodedata.normalize("NFC", str(self.name_forward))
        self.name_reverse = unicodedata.normalize("NFC", str(self.name_reverse))

        if (update_fields := kwargs.get("update_fields")) is not None:
            if "name_forward" in update_fields and "name_reverse" not in update_fields:
                modified_update_fields = set(update_fields)
                modified_update_fields.add("name_reverse")
                kwargs["update_fields"] = modified_update_fields

        super(Property, self).save(*args, **kwargs)
        return self


# TODO: comment and explain
def subj_or_obj_class_changed(sender, is_subj, **kwargs):
    def cascade_subj_obj_class_to_children(
        contenttype_to_add_or_remove,
        contenttype_already_saved_list,
        subj_or_obj_field_function,
    ):
        def get_all_parents(contenttype_current):
            parent_list = []
            class_current = contenttype_current.model_class()
            for class_parent in class_current.__bases__:
                # TODO: Avoid ContentType DB fetch
                contenttype_parent = ContentType.objects.filter(
                    model=class_parent.__name__
                )
                if len(contenttype_parent) == 1:
                    contenttype_parent = contenttype_parent[0]
                    parent_list.append(contenttype_parent)
                    parent_list.extend(get_all_parents(contenttype_parent))

            return parent_list

        def get_all_children(contenttype_current):
            child_list = []
            class_current = contenttype_current.model_class()
            for class_child in class_current.__subclasses__():
                # TODO: Avoid ContentType DB fetch
                contenttype_child = ContentType.objects.get_for_model(class_child)
                child_list.append(contenttype_child)
                child_list.extend(get_all_children(contenttype_child))

            return child_list

        parent_contenttype_list = get_all_parents(contenttype_to_add_or_remove)
        for parent_contenttype in parent_contenttype_list:
            if parent_contenttype in contenttype_already_saved_list:
                raise Exception(
                    f"Pre-existing parent class found when trying to save or remove a property subject or object class."
                    f" The current class to be saved is '{contenttype_to_add_or_remove.model_class().__name__}',"
                    f" but already saved is '{parent_contenttype.model_class().__name__}'."
                    f" Such a save could potentially be in conflict with an ontology."
                    f" Better save or remove the respective top parent subject or object class from this property."
                )
        children_contenttype_list = get_all_children(contenttype_to_add_or_remove)
        for child_contenttype in children_contenttype_list:
            subj_or_obj_field_function(child_contenttype)

    if kwargs["pk_set"] is not None and len(kwargs["pk_set"]) == 1:
        sending_property = kwargs["instance"]
        if sender == Property.subj_class.through:
            subj_or_obj_field = sending_property.subj_class
        elif sender == Property.obj_class.through:
            subj_or_obj_field = sending_property.obj_class
        else:
            raise Exception
        subj_or_obj_field_function = None
        if kwargs["action"] == "pre_add":
            subj_or_obj_field_function = subj_or_obj_field.add
        elif kwargs["action"] == "post_remove":
            subj_or_obj_field_function = subj_or_obj_field.remove
        if subj_or_obj_field_function is not None:
            cascade_subj_obj_class_to_children(
                contenttype_to_add_or_remove=ContentType.objects.get(
                    pk=min(kwargs["pk_set"])
                ),
                contenttype_already_saved_list=subj_or_obj_field.all(),
                subj_or_obj_field_function=subj_or_obj_field_function,
            )


def subj_class_changed(sender, **kwargs):
    subj_or_obj_class_changed(sender, is_subj=True, **kwargs)


def obj_class_changed(sender, **kwargs):
    subj_or_obj_class_changed(sender, is_subj=False, **kwargs)


m2m_changed.connect(subj_class_changed, sender=Property.subj_class.through)
m2m_changed.connect(obj_class_changed, sender=Property.obj_class.through)


class RelationPublishedQueryset(models.QuerySet):
    def filter_for_user(self, *args, **kwargs):
        if getattr(settings, "APIS_SHOW_ONLY_PUBLISHED", False):
            request = get_current_request()
            if request is not None:
                if request.user.is_authenticated:
                    return self
                else:
                    return self.filter(published=True)
            else:
                return self.filter(published=True)
        else:
            return self

    def filter_ann_proj(self, request=None, ann_proj=1, include_all=True):
        """The filter function provided by the manager class.

        :param request: `django.request` object
        :return: queryset that contains only objects that are shown in the highlighted text or those not connected
            to an annotation at all.
        """
        qs = self
        users_show = None
        if request:
            ann_proj = request.session.get("annotation_project", False)
            if not ann_proj:
                return qs
            users_show = request.session.get("users_show_highlighter", None)
        query = Q(annotation__annotation_project_id=ann_proj)
        if users_show is not None:
            query.add(Q(annotation__user_added_id__in=users_show), Q.AND)
        if include_all:
            query.add(Q(annotation__annotation_project__isnull=True), Q.OR)
        return qs.filter(query)


class InheritanceForwardManyToOneDescriptor(ForwardManyToOneDescriptor):
    def get_queryset(self, **hints):
        return self.field.remote_field.model.objects_inheritance.db_manager(
            hints=hints
        ).select_subclasses()


class InheritanceForeignKey(models.ForeignKey):
    forward_related_accessor_class = InheritanceForwardManyToOneDescriptor


class Triple(models.Model, GenericModel):
    subj = InheritanceForeignKey(
        RootObject,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="triple_set_from_subj",
        verbose_name="Subject",
    )
    obj = InheritanceForeignKey(
        RootObject,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="triple_set_from_obj",
        verbose_name="Object",
    )
    prop = models.ForeignKey(
        Property,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="triple_set_from_prop",
        verbose_name="Property",
    )

    objects = BaseRelationManager()
    objects_inheritance = InheritanceManager()

    def __repr__(self):
        try:
            return f"<{self.__class__.__name__}: subj: {self.subj}, prop: {self.prop}, obj: {self.obj}>"
        except RootObject.DoesNotExist:
            return f"<{self.__class__.__name__}: None>"

    def __str__(self):
        return self.__repr__()

    def get_web_object(self):
        return {
            "relation_pk": self.pk,
            "subj": str(self.subj),
            "obj": str(self.obj),
            "prop": self.prop.name_forward,
        }

    def save(self, *args, **kwargs):
        # TODO RDF: Integrate more proper check if subj and obj instances are of valid class as defined in prop.subj_class and prop.obj_class
        # def get_all_parents(cls_current):
        #     parent_list = []
        #     for p in cls_current.__bases__:
        #         parent_list.append(p)
        #         parent_list.extend(get_all_parents(p))
        #     return parent_list

        def get_all_childs(cls_current):
            child_list = []
            for p in cls_current.__subclasses__():
                child_list.append(p)
                child_list.extend(get_all_childs(p))

            return child_list

        if self.subj is None or self.obj is None or self.prop is None:
            raise Exception("subj, obj, or prop is None")
        if self.subj is not None:
            subj_class_name = self.subj.__class__.__name__
            if (
                ContentType.objects.get_for_model(self.subj.__class__)
                not in self.prop.subj_class.all()
            ):
                raise Exception(
                    f"Subject class '{subj_class_name}' is not in valid subject class list of property '{self.prop}'"
                )
        if self.obj is not None:
            obj_class_name = self.obj.__class__.__name__
            if (
                ContentType.objects.get_for_model(model=self.obj.__class__)
                not in self.prop.obj_class.all()
            ):
                raise Exception(
                    f"Object class '{obj_class_name}' is not in valid object class list of property '{self.prop}'"
                )

        super().save(*args, **kwargs)

    def duplicate(self):
        origin = self.__class__
        signals.pre_duplicate.send(sender=origin, instance=self)

        instance = copy.copy(self)

        self.pk = None
        self.id = None
        self._state.adding = True
        duplicate = self.save()

        signals.post_duplicate.send(
            sender=origin, instance=instance, duplicate=duplicate
        )
        return duplicate


class TempTriple(Triple, VersionMixin):
    review = models.BooleanField(
        default=False,
        help_text="Should be set to True, if the data record holds up quality standards.",
    )
    start_date = models.DateField(blank=True, null=True)
    start_start_date = models.DateField(blank=True, null=True)
    start_end_date = models.DateField(blank=True, null=True)
    end_date = models.DateField(blank=True, null=True)
    end_start_date = models.DateField(blank=True, null=True)
    end_end_date = models.DateField(blank=True, null=True)
    start_date_written = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name="Start",
    )
    end_date_written = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name="End",
    )
    # text = models.ManyToManyField("Text", blank=True)
    # collection = models.ManyToManyField("Collection")
    status = models.CharField(max_length=100)
    # source = models.ForeignKey(
    #     "Source", blank=True, null=True, on_delete=models.SET_NULL
    # )
    references = models.TextField(blank=True, null=True)
    notes = models.TextField(blank=True, null=True)

    def save(self, parse_dates=True, *args, **kwargs):
        """Adaption of the save() method of the class to automatically parse string-dates into date objects"""

        if parse_dates:
            # overwrite every field with None as default
            start_date = None
            start_start_date = None
            start_end_date = None
            end_date = None
            end_start_date = None
            end_end_date = None

            if self.start_date_written:
                # If some textual user input of a start date is there, then parse it

                start_date, start_start_date, start_end_date = DateParser.parse_date(
                    self.start_date_written
                )

            if self.end_date_written:
                # If some textual user input of an end date is there, then parse it

                end_date, end_start_date, end_end_date = DateParser.parse_date(
                    self.end_date_written
                )

            self.start_date = start_date
            self.start_start_date = start_start_date
            self.start_end_date = start_end_date
            self.end_date = end_date
            self.end_start_date = end_start_date
            self.end_end_date = end_end_date

        super().save(*args, **kwargs)

        return self



# Generated by Django 3.1.14 on 2023-01-24 12:01

import django.db.models.deletion
from django.db import migrations, models

import apis_core.apis_relations.models


class Migration(migrations.Migration):
    dependencies = [
        ("apis_metainfo", "0003_auto_20220218_1530"),
        ("apis_relations", "0003_auto_20230123_1715"),
    ]

    operations = [
        migrations.AlterField(
            model_name="property",
            name="name_forward",
            field=models.CharField(
                blank=True,
                help_text='Inverse relation like: "is sub-class of" vs. "is super-class of".',
                max_length=255,
                verbose_name="Name forward",
            ),
        ),
        migrations.AlterField(
            model_name="triple",
            name="obj",
            field=apis_core.apis_relations.models.InheritanceForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="triple_set_from_obj",
                to="apis_metainfo.rootobject",
                verbose_name="Object",
            ),
        ),
        migrations.AlterField(
            model_name="triple",
            name="prop",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="triple_set_from_prop",
                to="apis_relations.property",
                verbose_name="Property",
            ),
        ),
        migrations.AlterField(
            model_name="triple",
            name="subj",
            field=apis_core.apis_relations.models.InheritanceForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="triple_set_from_subj",
                to="apis_metainfo.rootobject",
                verbose_name="Subject",
            ),
        ),
    ]



import json

from django.forms.widgets import Input


class NewlineSeparatedListWidget(Input):
    input_type = "text"
    template_name = "widgets/multiline.html"

    class Media:
        js = ["js/multiline.js"]

    def value_from_datadict(self, data, files, name):
        values = [item for item in data.getlist(name) if item]
        return "\n".join(values)

    def format_value(self, value):
        if value == "" or value is None:
            return None
        return value.split("\n")


class JSONListWidget(Input):
    input_type = "text"
    template_name = "widgets/multiline.html"

    class Media:
        js = ["js/multiline.js"]

    def value_from_datadict(self, data, files, name):
        values = [item for item in data.getlist(name) if item]
        return values

    def format_value(self, value):
        if value == "" or value is None:
            return None
        return json.loads(value)



from django.contrib.contenttypes.models import ContentType
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.views.generic.base import TemplateView

from apis_core.generic.views import Create
from apis_core.relations.templatetags.relations import (
    possible_relation_types_from,
    relations_from,
)


class CreateRelation(Create):
    def get_form_kwargs(self, *args, **kwargs) -> dict:
        kwargs = super().get_form_kwargs(*args, **kwargs)
        kwargs["initial"]["obj_content_type"] = self.request.GET.get("obj_content_type")
        kwargs["initial"]["obj_object_id"] = self.request.GET.get("obj_object_id")
        kwargs["initial"]["subj_content_type"] = self.request.GET.get(
            "subj_content_type"
        )
        kwargs["initial"]["subj_object_id"] = self.request.GET.get("subj_object_id")
        return kwargs


class CreateRelationForm(CreateRelation):
    template_name = "relations/create_relation_form.html"

    def setup(self, *args, **kwargs) -> None:
        super().setup(*args, **kwargs)
        self.reverse = self.request.GET.get("reverse", "false").lower() == "true"

    def get(self, *args, **kwargs):
        resp = super().get(*args, **kwargs)
        content_type = ContentType.objects.get_for_model(self.model)
        resp["HX-Trigger-After-Settle"] = (
            '{"reinit_select2": "relation_' + content_type.model + '_form"}'
        )
        return resp

    def get_form_kwargs(self, *args, **kwargs) -> dict:
        kwargs = super().get_form_kwargs(*args, **kwargs)
        kwargs["reverse"] = self.reverse
        content_type = ContentType.objects.get_for_model(self.model)
        kwargs["hx_post_route"] = reverse(
            "apis_core:relations:create_relation_form", args=[content_type]
        )
        return kwargs

    def get_success_url(self) -> str:
        args = [
            self.object.obj_content_type,
            self.object.subj_content_type,
            self.object.subj_object_id,
        ]
        if self.reverse:
            args = [
                self.object.subj_content_type,
                self.object.obj_content_type,
                self.object.obj_object_id,
            ]
        return reverse("apis_core:relations:list_relations", args=args)


class ListRelations(TemplateView):
    template_name = "relations/list_relations.html"

    def get_context_data(self, *args, **kwargs) -> dict:
        ctx = super().get_context_data(*args, **kwargs)
        ctx["target"] = kwargs.get("target_contenttype")
        ctx["object"] = get_object_or_404(
            kwargs.get("object_contenttype").model_class(), pk=kwargs.get("object_id")
        )
        ctx["relations"] = relations_from(ctx["object"])
        ctx["possible_relations"] = possible_relation_types_from(ctx["object"])
        ctx["edit"] = True
        return ctx

    def get(self, *args, **kwargs):
        resp = super().get(*args, **kwargs)
        resp["HX-Trigger"] = "dismissModal"
        return resp



# Generated by Django 4.2.8 on 2023-12-22 07:56

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_vocabularies", "0004_delete_collectiontype"),
    ]

    operations = [
        migrations.DeleteModel(
            name="VocabsBaseClass",
        ),
    ]



import functools

from django.contrib.contenttypes.models import ContentType

from apis_core.relations.models import Relation


def is_relation(ct: ContentType) -> bool:
    mc = ct.model_class()
    return (
        issubclass(mc, Relation)
        and hasattr(mc, "subj_model")
        and hasattr(mc, "obj_model")
    )


@functools.cache
def relation_content_types(
    subj_model=None, obj_model=None, any_model=None, combination=(None, None)
) -> set[ContentType]:
    allcts = list(
        filter(
            lambda contenttype: contenttype.model_class() is not None,
            ContentType.objects.all(),
        )
    )
    relationcts = list(filter(lambda contenttype: is_relation(contenttype), allcts))
    if subj_model is not None:
        relationcts = list(
            filter(
                lambda contenttype: subj_model in contenttype.model_class().subj_list(),
                relationcts,
            )
        )
    if obj_model is not None:
        relationcts = list(
            filter(
                lambda contenttype: obj_model in contenttype.model_class().obj_list(),
                relationcts,
            )
        )
    if any_model is not None:
        relationcts = list(
            filter(
                lambda contenttype: any_model in contenttype.model_class().obj_list()
                or any_model in contenttype.model_class().subj_list(),
                relationcts,
            )
        )
    if all(combination):
        left, right = combination
        rels = list(
            filter(
                lambda contenttype: right in contenttype.model_class().obj_list()
                and left in contenttype.model_class().subj_list(),
                relationcts,
            )
        )
        rels.extend(
            list(
                filter(
                    lambda contenttype: left in contenttype.model_class().obj_list()
                    and right in contenttype.model_class().subj_list(),
                    relationcts,
                )
            )
        )
        relationcts = rels
    return set(relationcts)


def relation_match_target(relation, target: ContentType) -> bool:
    """
    test if a relation points to a target
    this function should not be cached, because the `forward` attribute
    is an annotation that does not seem to be part of the relation, so
    if cached, method could be called with another `forward` value and
    return the wrong result
    """
    if relation.forward and relation.obj_content_type == target:
        return True
    if not relation.forward and relation.subj_content_type == target:
        return True
    return False


@functools.cache
def get_all_relation_subj_and_obj() -> list[ContentType]:
    """
    Return the model classes of any model that is in some way
    connected to a relation - either as obj or as subj

    Returns:
    list[ContentType]: A list of unique ContentTypes for related models.
    """
    related_models = set()
    for rel in relation_content_types():
        related_models.update(rel.model_class().subj_list())
        related_models.update(rel.model_class().obj_list())
    return [ContentType.objects.get_for_model(item) for item in related_models]



from rest_framework import serializers

from apis_core.generic.serializers import GenericHyperlinkedIdentityField


class MinimalEntitySerializer(serializers.Serializer):
    uri = GenericHyperlinkedIdentityField(
        view_name="apis_core:generic:genericmodelapi-detail"
    )
    name = serializers.SerializerMethodField(method_name="get_name")

    def get_name(self, object) -> str:
        return str(object)



# Generated by Django 3.1.14 on 2022-02-01 22:49

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("apis_relations", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="property",
            name="property_class_uri",
            field=models.CharField(
                blank=True, max_length=255, verbose_name="Property Class URI"
            ),
        ),
    ]



# Generated by Django 4.2.8 on 2023-12-22 07:55

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_vocabularies", "0002_remove_texttype_collections_and_more"),
    ]

    operations = [
        migrations.DeleteModel(
            name="VocabNames",
        ),
        migrations.DeleteModel(
            name="VocabsUri",
        ),
    ]



from django.conf import settings
from django.contrib import admin
from django.urls import include, path
from drf_spectacular.views import (
    SpectacularAPIView,
    SpectacularRedocView,
    SpectacularSwaggerView,
)

from apis_core.generic.routers import CustomDefaultRouter

app_name = "apis_core"

urlpatterns = [
    path("", include("apis_core.core.urls")),
    path("", include("apis_core.generic.urls")),
    path("admin/", admin.site.urls),
    path("accounts/", include("django.contrib.auth.urls")),
]

router = CustomDefaultRouter()


if "apis_core.apis_metainfo" in settings.INSTALLED_APPS:
    from apis_core.apis_metainfo.urls import router as apis_metainfo_router

    router.registry.extend(apis_metainfo_router.registry)


if "apis_core.apis_entities" in settings.INSTALLED_APPS:
    urlpatterns.append(path("entities/", include("apis_core.apis_entities.urls")))
    from apis_core.apis_entities.urls import api_routes

    urlpatterns.append(path("api/", include(api_routes)))


if "apis_core.apis_relations" in settings.INSTALLED_APPS:
    urlpatterns.append(path("relations/", include("apis_core.apis_relations.urls")))


if "apis_core.relations" in settings.INSTALLED_APPS:
    urlpatterns.append(path("relations/", include("apis_core.relations.urls")))


if "apis_core.history" in settings.INSTALLED_APPS:
    urlpatterns.append(path("history/", include("apis_core.history.urls")))


if "apis_core.collections" in settings.INSTALLED_APPS:
    urlpatterns.append(path("collections/", include("apis_core.collections.urls")))


if "apis_core.documentation" in settings.INSTALLED_APPS:
    urlpatterns.append(path("", include("apis_core.documentation.urls")))


urlpatterns.append(path("api/", include(router.urls)))
urlpatterns.append(path("api-auth/", include("rest_framework.urls")))


urlpatterns.append(path("swagger/schema/", SpectacularAPIView.as_view(), name="schema"))
urlpatterns.append(
    path(
        "swagger/schema/swagger-ui/",
        SpectacularSwaggerView.as_view(url_name="apis_core:schema"),
        name="swagger-ui",
    )
)
urlpatterns.append(
    path(
        "swagger/schema/redoc/",
        SpectacularRedocView.as_view(url_name="apis_core:schema"),
        name="redoc",
    )
)



import django_filters
from django.contrib.contenttypes.models import ContentType
from django.db.models import Case, Q, Value, When
from django.db.models.functions import Concat

from apis_core.apis_entities.utils import get_entity_classes
from apis_core.apis_metainfo.models import RootObject
from apis_core.generic.helpers import generate_search_filter


def related_entity(queryset, name, value):
    from apis_core.apis_relations.models import Triple

    entities = get_entity_classes()
    q = Q()
    for entity in entities:
        name = entity._meta.model_name
        q |= Q(**{f"{name}__isnull": False}) & generate_search_filter(
            entity, value, prefix=f"{name}__"
        )
    all_entities = RootObject.objects_inheritance.filter(q).values_list("pk", flat=True)
    t = (
        Triple.objects.filter(Q(subj__in=all_entities) | Q(obj__in=all_entities))
        .annotate(
            related=Case(
                When(subj__in=all_entities, then="obj"),
                When(obj__in=all_entities, then="subj"),
            )
        )
        .values_list("related", flat=True)
    )
    return queryset.filter(pk__in=t)


class PropertyChoiceField(django_filters.fields.ModelChoiceField):
    def label_from_instance(self, obj):
        if obj.forward:
            targets = [ct.name for ct in obj.obj_class.all()]
        else:
            targets = [ct.name for ct in obj.subj_class.all()]
        return obj.name + " | " + ", ".join(targets)


class PropertyFilter(django_filters.ModelChoiceFilter):
    """
    A child of ModelChoiceFilter that only works with
    Properties, but in return it filters those so that
    only the Properties are listed that can be connected
    to the `model` given as argument.
    """

    field_class = PropertyChoiceField

    def __init__(self, *args, **kwargs):
        from apis_core.apis_relations.models import Property

        model = kwargs.pop("model", None)
        super().__init__(*args, **kwargs)

        if model is not None:
            ct = ContentType.objects.get_for_model(model)
            self.queryset = (
                Property.objects.all()
                .filter(Q(subj_class=ct) | Q(obj_class=ct))
                .annotate(
                    name=Case(
                        When(
                            obj_class=ct,
                            subj_class=ct,
                            then=Concat("name_forward", Value(" / "), "name_reverse"),
                        ),
                        When(obj_class=ct, then="name_reverse"),
                        When(subj_class=ct, then="name_forward"),
                    ),
                    forward=Case(
                        When(obj_class=ct, then=Value(False)),
                        When(subj_class=ct, then=Value(True)),
                    ),
                )
                .order_by("name")
                .distinct()
            )

    def filter(self, queryset, value):
        from apis_core.apis_relations.models import Property

        if value:
            p = Property.objects.get(name_forward=value)
            queryset = queryset.filter(triple_set_from_subj__prop=p).distinct()
        return queryset



import django_tables2 as tables
from django.conf import settings
from django.db.models import Case, F, When
from django.utils.html import format_html

from apis_core.apis_relations.models import TempTriple
from apis_core.generic.tables import GenericTable

empty_text_default = "There are currently no relations"


class TripleTable(GenericTable):
    subj = tables.Column(linkify=True)
    obj = tables.Column(linkify=True)

    class Meta(GenericTable.Meta):
        fields = ["id", "subj", "prop", "obj"]
        exclude = ["desc"]
        sequence = tuple(fields) + GenericTable.Meta.sequence


class SubjObjColumn(tables.ManyToManyColumn):
    def __init__(self, *args, **kwargs):
        kwargs["separator"] = format_html(",<br>")
        kwargs["orderable"] = True
        kwargs["filter"] = lambda qs: qs.order_by("model")
        kwargs["transform"] = lambda ent: f"{ent.name.capitalize()}"
        kwargs["linkify_item"] = lambda record: record.model_class().get_listview_url()
        super().__init__(*args, **kwargs)


class PropertyTable(GenericTable):
    """Construct table for properties.

    The table shows how entities connect with one another via properties (relations).
    It uses the format of an RDF triple – Subject-Predicate-Object – plus
    "Reverse Predicate" for the inverse relationship and is displayed on the frontend
    on the Relations > Property page.
    """

    # Note on constructing table columns / usage of variables:
    # The variables used to declare table columns need to have the same names
    # as the model field names from which the columns should be created,
    # or tables.Column needs to contain an attribute "accessor" which references
    # the original field name.
    # For columns which allow sorting, the variable names are used as sort strings
    # in the user's browser address bar, so for UX reasons, it may make sense to
    # use different variable names than the original field names.

    predicate = tables.Column(accessor="name_forward", verbose_name="Predicate")
    predicate_reverse = tables.Column(
        accessor="name_reverse", verbose_name="Reverse predicate"
    )
    subject = SubjObjColumn(accessor="subj_class", verbose_name="Subject")
    object = SubjObjColumn(accessor="obj_class", verbose_name="Object")

    class Meta(GenericTable.Meta):
        fields = ["subject", "predicate", "object", "predicate_reverse"]
        order_by = "predicate"
        exclude = ["desc"]
        sequence = tuple(fields) + GenericTable.Meta.sequence

    # Use order_ methods to define how individual columns should be sorted.
    # Method names are column names prefixed with "order_".
    # By default, columns for regular fields are sorted alphabetically; for
    # ManyToMany fields, however, the row IDs of the originating table are
    # used as basis for sorting.
    # When column names and field names differ (see earlier note), the original
    # field names need to be referenced when constructing queryset.
    def order_subject(self, queryset, is_descending):
        queryset = queryset.annotate(entity=F("subj_class__model")).order_by(
            ("-" if is_descending else "") + "entity"
        )
        return (queryset, True)

    def order_object(self, queryset, is_descending):
        queryset = queryset.annotate(entity=F("obj_class__model")).order_by(
            ("-" if is_descending else "") + "entity"
        )
        return (queryset, True)


class TripleTableBase(GenericTable):
    """
    The base table from which detail or edit tables will inherit from in order to avoid redundant definitions
    """

    class Meta:
        model = TempTriple

        # the fields list also serves as the defining order of them, as to avoid duplicated definitions
        fields = [
            "start_date_written",
            "end_date_written",
            "other_prop",
            "other_entity",
            "notes",
        ]
        exclude = (
            "desc",
            "view",
        )
        # reuse the list for ordering
        sequence = tuple(fields)

    def order_start_date_written(self, queryset, is_descending):
        if is_descending:
            return (queryset.order_by(F("start_date").desc(nulls_last=True)), True)
        return (queryset.order_by(F("start_date").asc(nulls_last=True)), True)

    def order_end_date_written(self, queryset, is_descending):
        if is_descending:
            return (queryset.order_by(F("end_date").desc(nulls_last=True)), True)
        return (queryset.order_by(F("end_date").asc(nulls_last=True)), True)

    def render_other_entity(self, record, value):
        """
        Custom render_FOO method for related entity linking. Since the 'other_related_entity' is a generated annotation
        on the queryset, it does not return the related instance but only the foreign key as the integer it is.
        Thus fetching the related instance is necessary.

        :param record: The 'row' of a queryset, i.e. an entity instance
        :param value: The current column of the row, i.e. the 'other_related_entity' annotation
        :return: related instance
        """

        if value == record.subj.pk:
            return record.subj

        elif value == record.obj.pk:
            return record.obj

        else:
            raise Exception(
                "Did not find the entity this relation is supposed to come from!"
                + "Something must have went wrong when annotating for the related instance."
            )

    def __init__(self, data, *args, **kwargs):
        data = data.annotate(
            other_entity=Case(
                # **kwargs pattern is needed here as the key-value pairs change with each relation class and entity instance.
                When(**{"subj__pk": self.entity_pk_self, "then": "obj"}),
                When(**{"obj__pk": self.entity_pk_self, "then": "subj"}),
            ),
            other_prop=Case(
                # **kwargs pattern is needed here as the key-value pairs change with each relation class and entity instance.
                When(**{"subj__pk": self.entity_pk_self, "then": "prop__name_forward"}),
                When(**{"obj__pk": self.entity_pk_self, "then": "prop__name_reverse"}),
            ),
        )

        self.base_columns["other_prop"].verbose_name = "Other property"
        self.base_columns[
            "other_entity"
        ].verbose_name = f"Related {self.other_entity_class_name.title()}"

        super().__init__(data, *args, **kwargs)

    def render_start_date_written(self, record, value):
        if record.start_start_date is not None and record.start_end_date is not None:
            title_text = f"{record.start_start_date} - {record.start_end_date}"
        elif record.start_date is not None:
            title_text = record.start_date
        else:
            return "—"
        return format_html(f"<abbr title='{title_text}'>{value}</b>")

    def render_end_date_written(self, record, value):
        if record.end_start_date is not None and record.end_end_date is not None:
            title_text = f"{record.end_start_date} - {record.end_end_date}"
        elif record.end_date is not None:
            title_text = record.end_date
        else:
            return "—"
        return format_html(f"<abbr title='{title_text}'>{value}</b>")


class TripleTableDetail(TripleTableBase):
    class Meta(TripleTableBase.Meta):
        exclude = TripleTableBase.Meta.exclude + ("delete", "edit")

    def __init__(self, data, *args, **kwargs):
        self.base_columns["other_entity"] = tables.Column(
            linkify=lambda record: record.obj.get_absolute_url()
            if record.other_entity == record.obj.id
            else record.subj.get_absolute_url()
        )

        # bibsonomy button
        if "apis_bibsonomy" in settings.INSTALLED_APPS:
            self.base_columns["ref"] = tables.TemplateColumn(
                template_name="apis_relations/references_button_generic_ajax_form.html"
            )

        super().__init__(data=data, *args, **kwargs)


class TripleTableEdit(TripleTableBase):
    class Meta(TripleTableBase.Meta):
        fields = TripleTableBase.Meta.fields
        if "apis_bibsonomy" in settings.INSTALLED_APPS:
            fields = ["ref"] + TripleTableBase.Meta.fields
        sequence = tuple(fields)

    def __init__(self, *args, **kwargs):
        self.base_columns["other_entity"] = tables.Column(
            linkify=lambda record: record.obj.get_edit_url()
            if record.other_entity == record.obj.id
            else record.subj.get_edit_url()
        )

        self.base_columns["edit"] = tables.TemplateColumn(
            template_name="apis_relations/edit_button_generic_ajax_form.html"
        )

        if "apis_bibsonomy" in settings.INSTALLED_APPS:
            self.base_columns["ref"] = tables.TemplateColumn(
                template_name="apis_relations/references_button_generic_ajax_form.html"
            )

        super().__init__(*args, **kwargs)


def get_generic_triple_table(other_entity_class_name, entity_pk_self, detail):
    if detail:
        tt = TripleTableDetail
    else:
        tt = TripleTableEdit
    tt.entity_pk_self = entity_pk_self
    tt.other_entity_class_name = other_entity_class_name
    return tt



from django.urls import path, re_path

from apis_core.apis_entities.autocomplete3 import PropertyAutocomplete

from . import views

app_name = "apis_relations"

urlpatterns = [
    path("ajax/get/", views.get_form_ajax, name="get_form_ajax"),
    re_path(
        r"^ajax/save/(?P<entity_type>\w+)/(?P<kind_form>\w+)/(?P<SiteID>[0-9]+)(?:/(?P<ObjectID>[0-9]*))?/$",
        views.save_ajax_form,
        name="save_ajax_form",
    ),
    re_path(
        r"^(?P<entity>[a-z0-9_]+)/list/$",
        views.GenericRelationView.as_view(),
        name="generic_relations_list",
    ),
    re_path(
        r"^autocomplete/(?P<entity_self>[a-zA-Z0-9-_]+)/(?P<entity_other>[a-zA-Z0-9-_]+)/$",
        PropertyAutocomplete.as_view(),
        name="generic_property_autocomplete",
    ),
]



#!/usr/bin/python
# -*- coding: utf-8 -*-
import json

from dal import autocomplete
from django import http
from django.contrib.contenttypes.models import ContentType

from apis_core.apis_entities.utils import get_entity_classes
from apis_core.apis_relations.models import Property


class PropertyAutocomplete(autocomplete.Select2ListView):
    # These constants are set so that they are defined in one place only and reused by fetching them elsewhere.
    SELF_SUBJ_OTHER_OBJ_STR = "self_subj_other_obj"
    SELF_OBJ_OTHER_SUBJ_STR = "self_obj_other_subj"

    def get_autocomplete_property_choices(self, self_model, other_model, search_str):
        contenttypes = [
            ContentType.objects.get_for_model(model) for model in get_entity_classes()
        ]
        self_contenttype = next(
            filter(lambda x: x.model == self_model.lower(), contenttypes)
        )
        other_contenttype = next(
            filter(lambda x: x.model == other_model.lower(), contenttypes)
        )

        rbc_self_subj_other_obj = Property.objects.filter(
            subj_class=self_contenttype,
            obj_class=other_contenttype,
            name_forward__icontains=search_str,
        )
        rbc_self_obj_other_subj = Property.objects.filter(
            subj_class=other_contenttype,
            obj_class=self_contenttype,
            name_reverse__icontains=search_str,
        )
        choices = []
        for rbc in rbc_self_subj_other_obj:
            choices.append(
                {
                    # misuse of the id item as explained above
                    "id": f"id:{rbc.pk}__direction:{self.SELF_SUBJ_OTHER_OBJ_STR}",
                    "text": rbc.name_forward,
                }
            )
        for rbc in rbc_self_obj_other_subj:
            choices.append(
                {
                    # misuse of the id item as explained above
                    "id": f"id:{rbc.pk}__direction:{self.SELF_OBJ_OTHER_SUBJ_STR}",
                    "text": rbc.name_reverse,
                }
            )
        return choices

    def get(self, request, *args, **kwargs):
        more = False
        choices = self.get_autocomplete_property_choices(
            kwargs["entity_self"], kwargs["entity_other"], self.q
        )
        return http.HttpResponse(
            json.dumps(
                {"results": choices, "pagination": {"more": more}, "abcde": "0"}
            ),
            content_type="application/json",
        )



# Generated by Django 3.1.14 on 2022-02-01 12:41

import django.db.models.deletion
from django.db import migrations, models

import apis_core.apis_relations.models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("apis_metainfo", "0002_auto_20220201_1241"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Property",
            fields=[
                (
                    "rootobject_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="apis_metainfo.rootobject",
                    ),
                ),
                (
                    "name_forward",
                    models.CharField(
                        blank=True,
                        help_text='Inverse relation like: "is sub-class of" vs. "is super-class of".',
                        max_length=255,
                        verbose_name="Name reverse",
                    ),
                ),
                (
                    "name_reverse",
                    models.CharField(
                        blank=True,
                        help_text='Inverse relation like: "is sub-class of" vs. "is super-class of".',
                        max_length=255,
                        verbose_name="Name reverse",
                    ),
                ),
                (
                    "obj_class",
                    models.ManyToManyField(
                        limit_choices_to=models.Q(app_label="apis_entities"),
                        related_name="property_set_obj",
                        to="contenttypes.ContentType",
                    ),
                ),
                (
                    "subj_class",
                    models.ManyToManyField(
                        limit_choices_to=models.Q(app_label="apis_entities"),
                        related_name="property_set_subj",
                        to="contenttypes.ContentType",
                    ),
                ),
            ],
            bases=("apis_metainfo.rootobject",),
        ),
        migrations.CreateModel(
            name="Triple",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "obj",
                    apis_core.apis_relations.models.InheritanceForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="triple_set_from_obj",
                        to="apis_metainfo.rootobject",
                    ),
                ),
                (
                    "prop",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="triple_set_from_prop",
                        to="apis_relations.property",
                    ),
                ),
                (
                    "subj",
                    apis_core.apis_relations.models.InheritanceForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="triple_set_from_subj",
                        to="apis_metainfo.rootobject",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="TempTriple",
            fields=[
                (
                    "triple_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="apis_relations.triple",
                    ),
                ),
                (
                    "review",
                    models.BooleanField(
                        default=False,
                        help_text="Should be set to True, if the data record holds up quality standards.",
                    ),
                ),
                ("start_date", models.DateField(blank=True, null=True)),
                ("start_start_date", models.DateField(blank=True, null=True)),
                ("start_end_date", models.DateField(blank=True, null=True)),
                ("end_date", models.DateField(blank=True, null=True)),
                ("end_start_date", models.DateField(blank=True, null=True)),
                ("end_end_date", models.DateField(blank=True, null=True)),
                (
                    "start_date_written",
                    models.CharField(
                        blank=True, max_length=255, null=True, verbose_name="Start"
                    ),
                ),
                (
                    "end_date_written",
                    models.CharField(
                        blank=True, max_length=255, null=True, verbose_name="End"
                    ),
                ),
                ("status", models.CharField(max_length=100)),
                ("references", models.TextField(blank=True, null=True)),
                ("notes", models.TextField(blank=True, null=True)),
            ],
            bases=("apis_relations.triple",),
        ),
    ]



# SPDX-FileCopyrightText: 2023 Birger Schacht
# SPDX-License-Identifier: MIT

import importlib
import logging
import re
from collections import defaultdict
from typing import Tuple

from AcdhArcheAssets.uri_norm_rules import get_normalized_uri
from rdflib import Graph

from apis_core.utils.settings import dict_from_toml_directory

logger = logging.getLogger(__name__)


def definition_matches_model(definition: str, model: object) -> bool:
    if definition.get("superclass", False) and model:
        try:
            module, cls = definition.get("superclass").rsplit(".", 1)
            module = importlib.import_module(module)
            parent = getattr(module, cls)
            if issubclass(type(model), parent) or issubclass(model, parent):
                return True
        except Exception as e:
            logger.error("superclass %s led to: %s", definition.get("superclass"), e)
    return False


def definition_matches_uri(definition: str, uri: str) -> bool:
    if regex := definition.get("regex", False):
        logger.info("found regex %s", regex)
        pattern = re.compile(regex)
        if pattern.fullmatch(uri) is None:
            return False
    return True


def get_definition_and_attributes_from_uri(
    uri: str, model: object
) -> Tuple[dict, dict]:
    """
    This function looks for `.toml` files in the `rdfimport` app directories
    and loads all the files it can parse. For every file that contains a
    `superclass` key it checks if it is a superclass of `model`.
    It uses the first file that
    matches to extract attributes from the RDF endpoint and then returns both
    the parsed file contents and the extracted attributes.
    The reason we are also returning the parsed file contents is, that you then
    can define a model *in* the file and then use this function to iterate over
    a list of URIs and you can use the matched definition to choose which model
    to create.
    The dict containing the parsed file contents also contains the filename, to
    make debugging a bit easier.
    """
    uri = get_normalized_uri(uri)
    graph = Graph()
    graph.parse(uri)

    configs = dict_from_toml_directory("rdfimport")
    matching_definition = None
    for key, definition in configs.items():
        if definition_matches_model(definition, model) and definition_matches_uri(
            definition, uri
        ):
            matching_definition = definition
            matching_definition["filename"] = str(key)
            break
    model_attributes = defaultdict(list)
    if matching_definition:
        attributes = matching_definition.get("attributes", [])
        sparql_attributes = list(filter(lambda d: d.get("sparql"), attributes))
        for attribute in sparql_attributes:
            result = graph.query(attribute["sparql"])
            for binding in result.bindings:
                # {rdflib.term.Variable('somekey'): rdflib.term.Literal('some value')}
                for key, value in binding.items():
                    model_attributes[str(key)].append(str(value))
    else:
        raise AttributeError(f"No matching definition found for {uri}")
    return matching_definition, model_attributes



# Generated by Django 3.1.14 on 2022-02-01 12:41

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("apis_metainfo", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="TempEntityClass",
            fields=[
                (
                    "rootobject_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="apis_metainfo.rootobject",
                    ),
                ),
                (
                    "review",
                    models.BooleanField(
                        default=False,
                        help_text="Should be set to True, if the data record holds up quality standards.",
                    ),
                ),
                ("start_date", models.DateField(blank=True, null=True)),
                ("start_start_date", models.DateField(blank=True, null=True)),
                ("start_end_date", models.DateField(blank=True, null=True)),
                ("end_date", models.DateField(blank=True, null=True)),
                ("end_start_date", models.DateField(blank=True, null=True)),
                ("end_end_date", models.DateField(blank=True, null=True)),
                (
                    "start_date_written",
                    models.CharField(
                        blank=True, max_length=255, null=True, verbose_name="Start"
                    ),
                ),
                (
                    "end_date_written",
                    models.CharField(
                        blank=True, max_length=255, null=True, verbose_name="End"
                    ),
                ),
                ("status", models.CharField(max_length=100)),
                ("references", models.TextField(blank=True, null=True)),
                ("notes", models.TextField(blank=True, null=True)),
                ("published", models.BooleanField(default=False)),
                ("collection", models.ManyToManyField(to="apis_metainfo.Collection")),
                (
                    "source",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="apis_metainfo.source",
                    ),
                ),
                ("text", models.ManyToManyField(blank=True, to="apis_metainfo.Text")),
            ],
            options={
                "abstract": False,
            },
            bases=("apis_metainfo.rootobject",),
        ),
    ]



from django import template

from apis_core.apis_relations.utils import triple_sidebar

register = template.Library()


@register.simple_tag(takes_context=True)
def object_relations(context, detail=True):
    obj = context["object"]
    return triple_sidebar(obj, context["request"], detail)



from urllib.parse import urlsplit

from AcdhArcheAssets.uri_norm_rules import get_normalized_uri
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import models
from model_utils.managers import InheritanceManager

from apis_core.generic.abc import GenericModel
from apis_core.utils import rdf
from apis_core.utils import settings as apis_settings


class RootObject(GenericModel, models.Model):
    """
    The very root thing that can exist in a given ontology. Several classes inherit from it.
    By having one overarching super class we gain the advantage of unique identifiers.
    """

    # self_contenttype: a foreign key to the respective contenttype comes in handy when querying for
    # triples where the subject's or object's contenttype must be respected (e.g. get all triples
    # where the subject is a Person)
    self_contenttype = models.ForeignKey(
        ContentType,
        on_delete=models.deletion.CASCADE,
        null=True,
        blank=True,
        editable=False,
    )
    objects = models.Manager()
    objects_inheritance = InheritanceManager()

    def save(self, *args, **kwargs):
        self.self_contenttype = ContentType.objects.get_for_model(self)
        super().save(*args, **kwargs)


# Uri model
# We use a custom UriManager, so we can override the queryset `get`
# method. This way we can normalize the uri field.


class UriQuerySet(models.query.QuerySet):
    def get(self, *args, **kwargs):
        if "uri" in kwargs:
            kwargs["uri"] = get_normalized_uri(kwargs["uri"])
        return super().get(*args, **kwargs)


class UriManager(models.Manager):
    def get_queryset(self):
        return UriQuerySet(self.model)


class Uri(GenericModel, models.Model):
    uri = models.URLField(blank=True, null=True, unique=True, max_length=255)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True)
    object_id = models.PositiveIntegerField(null=True)
    content_object = GenericForeignKey()

    objects = UriManager()

    def __str__(self):
        return str(self.uri)

    def get_web_object(self):
        result = {
            "relation_pk": self.pk,
            "relation_type": "uri",
            "related_root_object": self.content_object.name,
            "related_root_object_url": self.content_object.get_absolute_url(),
            "related_root_object_class_name": self.content_object.__class__.__name__.lower(),
            "uri": self.uri,
        }
        return result

    def save(self, *args, **kwargs):
        self.clean()
        return super().save(*args, **kwargs)

    def clean(self):
        self.uri = get_normalized_uri(self.uri)
        if self.uri and not hasattr(self, "content_object"):
            try:
                definition, attributes = rdf.get_definition_and_attributes_from_uri(
                    self.uri
                )
                if definition.getattr("model", False) and attributes:
                    app_label, model = definition.getattr("model").split(".", 1)
                    ct = ContentType.objects.get_by_natural_key(app_label, model)
                    obj = ct.model_class()(**attributes)
                    obj.save()
                    self.content_type = ContentType.objects.get_for_model(obj)
                    self.object_id = obj.id
                else:
                    raise ImproperlyConfigured(
                        f"{self.uri}: did not find matching rdf defintion"
                    )
            except Exception as e:
                raise ValidationError(f"{e}: {self.uri}")

    def internal(self) -> bool:
        my_netloc = urlsplit(self.uri).netloc
        return any(
            [my_netloc == urlsplit(uri).netloc for uri in apis_settings.internal_uris()]
        )



from django.core.exceptions import ValidationError
from django.forms import ChoiceField, ModelChoiceField, MultiValueField, MultiWidget
from django.utils.translation import gettext as _

from apis_core.utils.helpers import create_object_from_uri


class ModelImportChoiceField(ModelChoiceField):
    def to_python(self, value):
        result = None
        try:
            result = create_object_from_uri(value, self.queryset.model)
        except Exception as e:
            raise ValidationError(
                _("Could not import %(value)s: %(exception)s"),
                params={"value": value, "exception": e},
            )
        return result or super().to_python(value)


class IncludeExcludeMultiWidget(MultiWidget):
    template_name = "widgets/includeexclude_multiwidget.html"
    use_fieldset = False

    def decompress(self, value):
        return [value, value]


class IncludeExcludeField(MultiValueField):
    """
    This is a custom MultiValueField that adds a ChoiceField that only provides two
    choices, namely `exclude` and `include`. It can be used for django-filter filters
    to specify which action should be done with the filter.
    """

    def __init__(self, field, *args, **kwargs):
        fields = (
            field,
            ChoiceField(choices=[("include", "include"), ("exclude", "exclude")]),
        )
        kwargs["widget"] = IncludeExcludeMultiWidget(widgets=[f.widget for f in fields])
        super().__init__(fields=fields, *args, **kwargs)

    def compress(self, data_list):
        return data_list



import itertools

from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django_tables2 import RequestConfig

from apis_core.apis_relations.models import Property, TempTriple
from apis_core.apis_relations.tables import get_generic_triple_table
from apis_core.utils.settings import get_entity_settings_by_modelname


def get_content_types_with_allowed_relation_from(
    content_type: ContentType,
) -> list[ContentType]:
    """Returns a list of ContentTypes to which the given ContentTypes may be related by a Property"""

    # Find all the properties where the entity is either subject or object
    properties_with_entity_as_subject = Property.objects.filter(
        subj_class=content_type
    ).prefetch_related("obj_class")
    properties_with_entity_as_object = Property.objects.filter(
        obj_class=content_type
    ).prefetch_related("subj_class")

    content_type_querysets = []

    # Where entity is subject, get all the object content_types
    for p in properties_with_entity_as_subject:
        objs = p.obj_class.all()
        content_type_querysets.append(objs)
    # Where entity is object, get all the subject content_types
    for p in properties_with_entity_as_object:
        subjs = p.subj_class.all()
        content_type_querysets.append(subjs)

    # Join querysets with itertools.chain, call set to make unique, and extract the model class
    return set(itertools.chain(*content_type_querysets))


def triple_sidebar(obj: object, request, detail=True):
    content_type = ContentType.objects.get_for_model(obj)
    side_bar = []

    triples_related_all = (
        TempTriple.objects_inheritance.filter(Q(subj__pk=obj.pk) | Q(obj__pk=obj.pk))
        .all()
        .select_subclasses()
    )

    for other_content_type in get_content_types_with_allowed_relation_from(
        content_type
    ):
        triples_related_by_entity = triples_related_all.filter(
            (Q(subj__self_contenttype=other_content_type) & Q(obj__pk=obj.pk))
            | (Q(obj__self_contenttype=other_content_type) & Q(subj__pk=obj.pk))
        )

        table_class = get_generic_triple_table(
            other_entity_class_name=other_content_type.model,
            entity_pk_self=obj.pk,
            detail=detail,
        )

        prefix = f"{other_content_type.model}"
        title_card = other_content_type.name
        tb_object = table_class(data=triples_related_by_entity, prefix=prefix)
        tb_object_open = request.GET.get(prefix + "page", None)
        entity_settings = get_entity_settings_by_modelname(content_type.model)
        per_page = entity_settings.get("relations_per_page", 10)
        RequestConfig(request, paginate={"per_page": per_page}).configure(tb_object)
        tab_id = f"triple_form_{content_type.model}_to_{other_content_type.model}"
        side_bar.append(
            (
                title_card,
                tb_object,
                tab_id,
                tb_object_open,
            )
        )
    return side_bar



# Generated by Django 4.2.8 on 2023-12-22 07:29

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_metainfo", "0008_alter_rootobject_self_contenttype"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="collection",
            name="collection_type",
        ),
    ]



from django_tables2 import TemplateColumn

from apis_core.generic.tables import GenericTable


class SkosCollectionContentObjectTable(GenericTable):
    target = TemplateColumn(
        "<a href='{{ record.content_object.get_absolute_url }}'>{{ record.content_object }}</a>",
        orderable=False,
    )



import json

from django.contrib.auth.decorators import login_required
from django.contrib.contenttypes.models import ContentType
from django.http import HttpResponse
from django.template.loader import render_to_string

from apis_core.apis_entities.autocomplete3 import PropertyAutocomplete
from apis_core.apis_entities.utils import get_entity_classes
from apis_core.apis_relations.forms import GenericTripleForm
from apis_core.apis_relations.models import Property, TempTriple, Triple
from apis_core.generic.views import List


class GenericRelationView(List):
    def setup(self, *args, **kwargs):
        super().setup(*args, **kwargs)
        if self.kwargs["entity"] == "property":
            self.model = Property
        else:
            self.model = Triple
        self.queryset = self.model.objects.all()


# TODO RDF: After full conversion to ne ajax logic, remove this function
@login_required
def get_form_ajax(request):
    """Returns forms rendered in html"""

    form_name = request.POST.get("FormName")
    SiteID = int(request.POST.get("SiteID"))
    ButtonText = request.POST.get("ButtonText")
    ObjectID = request.POST.get("ObjectID")
    entity_type_self_str = request.POST.get("entity_type")

    if ObjectID is None and form_name.startswith("triple_form_"):
        # If this is the case, then instantiate an empty form

        entity_type_other_str = form_name.split("_to_")[1]

        form = GenericTripleForm(
            entity_type_self_str=entity_type_self_str,
            entity_type_other_str=entity_type_other_str,
        )

    elif ObjectID is not None and SiteID is not None:
        # If this is the case, then instantiate a form and pre-load with existing data

        triple = TempTriple.objects.get(pk=ObjectID)
        property_instance = triple.prop

        if triple.subj.pk == SiteID:
            entity_instance_self = triple.subj
            entity_instance_other = triple.obj
            property_direction = PropertyAutocomplete.SELF_SUBJ_OTHER_OBJ_STR
        elif triple.obj.pk == SiteID:
            entity_instance_self = triple.obj
            entity_instance_other = triple.subj
            property_direction = PropertyAutocomplete.SELF_OBJ_OTHER_SUBJ_STR
        else:
            raise Exception("SiteID was not found in triple")

        entity_type_other_str = entity_instance_other.__class__.__name__
        entity_type_self_str = entity_instance_self.__class__.__name__
        form_name = f"triple_form_{entity_type_self_str.lower()}_to_{entity_type_other_str.lower()}"

        form = GenericTripleForm(
            entity_type_self_str=entity_type_self_str,
            entity_type_other_str=entity_type_other_str,
        )
        form.load_subj_obj_prop(
            entity_instance_self,
            entity_instance_other,
            property_instance,
            property_direction,
        )
        form.load_remaining_data_from_triple(triple)

    else:
        raise Exception("Missing necessary form data")

    param_dict = {
        "entity_type": entity_type_self_str,
        "form": form,
        "type1": form_name,
        "url2": "save_ajax_" + form_name,
        "button_text": ButtonText,
        "ObjectID": ObjectID,
        "SiteID": SiteID,
    }

    rendered_form_str = render_to_string(
        "apis_relations/_ajax_form.html",  # rel_form_logic_breadcrumb (for refinding the implicit connections)
        context=param_dict,
    )

    data = {"tab": form_name, "form": rendered_form_str}

    return HttpResponse(json.dumps(data), content_type="application/json")


# TODO RDF: Re-implement highlighter and label form
@login_required
def save_ajax_form(
    request, entity_type, kind_form, SiteID, ObjectID=False
):  # rel_form_logic_breadcrumb (for refinding the implicit connections)
    """Tests validity and saves AjaxForms, returns them when validity test fails"""

    self_other = kind_form.split("triple_form_")[1].split("_to_")
    entity_type_self_str = self_other[0]
    entity_type_other_str = self_other[1]
    contenttypes = [
        ContentType.objects.get_for_model(model) for model in get_entity_classes()
    ]
    entity_type_self_class = next(
        filter(lambda x: x.model == entity_type_self_str.lower(), contenttypes)
    ).model_class()
    entity_type_other_class = next(
        filter(lambda x: x.model == entity_type_other_str.lower(), contenttypes)
    ).model_class()
    entity_instance_self = entity_type_self_class.objects.get(pk=SiteID)
    entity_instance_other = entity_type_other_class.get_or_create_uri(
        uri=request.POST["other_entity"]
    )
    start_date_written = request.POST["start_date_written"]
    end_date_written = request.POST["end_date_written"]
    property_param_dict = {}
    for param_pair in request.POST["property"].split("__"):
        param_pair_split = param_pair.split(":")
        property_param_dict[param_pair_split[0]] = param_pair_split[1]
    property_instance = Property.objects.get(pk=property_param_dict["id"])
    property_direction = property_param_dict["direction"]

    form = GenericTripleForm(entity_type_self_str, entity_type_other_str)
    form.load_subj_obj_prop(
        entity_instance_self,
        entity_instance_other,
        property_instance,
        property_direction,
    )

    if ObjectID is not False:
        triple = TempTriple.objects.get(pk=ObjectID)
        form.load_remaining_data_from_triple(triple)

    form.load_remaining_data_from_input(
        start_date_written,
        end_date_written,
    )
    form.fields["notes"].initial = request.POST["notes"]
    form.save()

    data = {
        "test": True,
        "tab": kind_form,
        "call_function": "EntityRelationForm_response",
        "instance": form.instance.get_web_object(),
        "table_html": form.get_html_table(
            entity_instance_self, entity_instance_other, request
        ).as_html(request),
        "text": None,
        "right_card": True,
    }
    return HttpResponse(json.dumps(data), content_type="application/json")



from dal import autocomplete
from django.contrib import messages
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.shortcuts import get_object_or_404, redirect
from django.utils.html import format_html
from django.views import View

from apis_core.apis_metainfo.models import RootObject
from apis_core.generic.helpers import generate_search_filter
from apis_core.generic.views import GenericModelMixin


class EntitiesDuplicate(GenericModelMixin, PermissionRequiredMixin, View):
    permission_action_required = "create"

    def get(self, request, *args, **kwargs):
        source_obj = get_object_or_404(self.model, pk=kwargs["pk"])
        newobj = source_obj.duplicate()

        messages.success(
            request,
            format_html(
                "<a href={}>{}</a> was successfully duplicated to the current object:",
                source_obj.get_absolute_url(),
                source_obj,
            ),
        )
        return redirect(newobj.get_edit_url())


class EntitiesAutocomplete(autocomplete.Select2QuerySetView):
    """
    This endpoint allows us to use autocomplete over multiple model classes.
    It takes a parameter `entities` which is a list of ContentType natural
    keys and searches for the query in all instances of those entities
    (using `generate_search_filter`, which means it uses a different search
    approach for every model).
    The return values of the endpoint are then prefixed with the id of the
    contenttype of the results, separated by an underscore.

    Example:
    Using this endpoint with the parameters:

        ?entities=apis_ontology.person&entities=apis_ontology.place&q=ammer

    gives you all the persons and places that have `ammer` in their names
    and labels.
    """

    def get_result_value(self, result) -> str:
        content_type = ContentType.objects.get_for_model(result)
        return f"{content_type.id}_" + super().get_result_value(result)

    def get_queryset(self):
        q = Q()
        if entities := self.request.GET.getlist("entities"):
            for entity in entities:
                app_label, model = entity.split(".")
                content_type = get_object_or_404(
                    ContentType, app_label=app_label, model=model
                )
                name = (
                    RootObject.objects_inheritance.get_queryset()._get_ancestors_path(
                        content_type.model_class()
                    )
                )
                q |= Q(**{f"{name}__isnull": False}) & generate_search_filter(
                    content_type.model_class(), self.q, prefix=f"{name}__"
                )
        return RootObject.objects_inheritance.select_subclasses().filter(q)



from rest_framework.authentication import TokenAuthentication


# not used in APIS, but can be used in downstream projects
class TokenAuthSupportQueryString(TokenAuthentication):
    """
    Extend the TokenAuthentication class to support querystring authentication
    in the form of "http://www.example.com/?auth_token=<api_key>"
    """

    def authenticate(self, request):
        # Check if 'token_auth' is in the request query params.
        # Give precedence to 'Authorization' header.
        if (
            "api_key" in request.query_params
            and "HTTP_AUTHORIZATION" not in request.META
        ):
            return self.authenticate_credentials(request.query_params.get("api_key"))
        else:
            return super().authenticate(request)



from django.urls import path

from . import views

app_name = "collections"

urlpatterns = [
    path(
        "collectionobjecttoggle/<int:content_type_id>/<int:object_id>/<int:collection>",
        views.CollectionToggle.as_view(),
        name="collectiontoggle",
    ),
    path(
        "collectionobjectparent/<int:content_type_id>/<int:object_id>/<int:collectionobject>",
        views.CollectionObjectParent.as_view(),
        name="collectionobjectparent",
    ),
    path(
        "collectionsessiontoggle/<int:skoscollection>",
        views.CollectionSessionToggle.as_view(),
        name="collectionsessiontoggle",
    ),
]
